# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import logging


import vulkn.sql.utils as utils
import vulkn.sql.library as library
import vulkn.session as session
from vulkn.contrib.sqlparse import sqlparse


log = logging.getLogger()


class SQLStatement:
    def __init__(self, statement: str) -> None:
        self._original_statement = statement.strip(';').strip()

    def __str__(self) -> str:
        return self._original_statement

    def ASTAnalyzeRewrite(self) -> str:
        ast = sqlparse.parse(self._original_statement)[0]
        if ast.get_type() == 'SELECT':
            for ext in library.extensions.keys():
                ast = library.extensions[ext](ast)
        return ast

    def optimize(self):
        if session.enable_parser:
            return str(self.ASTAnalyzeRewrite())
        return self._original_statement


# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import logging


import vulkn.sql.library as vlib
from vulkn.contrib.sqlparse import sqlparse


log = logging.getLogger()
__priority__ = 1


def ASTRewriteUserDefinedFunctions(ast):
    ret = sqlparse.parse(str(sqlparse.sql.TokenList(list(_rewrite(ast)))))[0]
    return ret

def _rewrite(ast):
    tree = ast.tokens if isinstance(ast, sqlparse.sql.TokenList) else ast
    for idx, token in enumerate(tree):
        if isinstance(token, sqlparse.sql.Parenthesis):
            if len(token.tokens) > 1 or not isinstance(token.tokens[0], sqlparse.sql.Token):
                for t in _rewrite(token):
                    yield t
            else:
                yield token
        elif isinstance(token, sqlparse.sql.Function):
            if token.tokens[0].value in vlib.udfs.keys():
                parameters = _rewrite(token.get_parameters())
                yield sqlparse.sql.Token(
                    sqlparse.tokens.Literal,
                    vlib.udfs[token.tokens[0].value](*list(parameters)))
            else:
                yield token
        elif isinstance(token, sqlparse.sql.Identifier):
            if len(token.tokens) > 1 or not isinstance(token.tokens[0], sqlparse.sql.Token):
                for t in _rewrite(token):
                    yield t
            else:
                yield token
        elif isinstance(token, sqlparse.sql.IdentifierList):
            for t in _rewrite(token):
                yield t
        else:
            yield token



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import csv
import os
import ast
import re


import vulkn
from vulkn.clickhouse import sqlformat
from vulkn.formats.csv import CSV
from vulkn.storages.posix import PosixStorage
from vulkn.datatable import BaseTableDataTable
from vulkn.utils import timer


class Reader:
    def __init__(self):
        pass

    def format(self, format):
        # TODO: Next release. Auto-discover/load formats
        if format == 'csv':
            return DataManager(CSV())


class DataManager:
    def __init__(self, format):
        self._format = format

    def options(self, **kwargs):
        self._format = self._format.options(**kwargs)
        return self

    @timer
    def load(self, uri, database, table):
        # TODO: Next release. Use proper vulkn method
        def _create_table(database, table, columns, engine):
            ctx = vulkn.Vulkn()
            column_ddl = ', '.join(['"{}" {}'.format(k, v) for k, v in columns.items()])
            create_ddl = f'CREATE TABLE "{database}"."{table}" ({column_ddl}) ENGINE = {engine}'
            if not ctx.exec(create_ddl):
                raise Exception('Unable to create table')

        schema = self._format._options['schema']
        tmp_db = database
        ctx = vulkn.Vulkn()

        if schema:
            cols = [c._name for c in schema]
            types = [c._col_type for c in schema]

        if schema is None or self._format._options['infer_schema']:
            cols = self._format.columns(PosixStorage(uri))
            types = ['String']*len(cols)
            if self._format._options['column_format'] == 'snake_case':
                cols = [snake_case(c) for c in cols]
            sample_engine = self._format._options['sample_engine'] or vulkn.engines.Memory()
            ctx.dropTable(tmp_db, f'tmp_{table}')
            _create_table(tmp_db, f'tmp_{table}', dict(zip(cols, types)), sample_engine)
            # TODO: Next release. Use storage.write method
            sample = self._format.sample(PosixStorage(uri))
            ctx._conn.insert_blob(sample, tmp_db + f'."tmp_{table}"', 'CSV')
            types = infer_column_types(tmp_db,
                                       f'tmp_{table}',
                                       cols,
                                       self._format._options['sample_size'],
                                       self._format._options['allow_enums'])
            ctx.dropTable(tmp_db, f'tmp_{table}')

        if self._format._options['overwrite']:
            ctx.dropTable(database, table)

        _create_table(database,
                      table,
                      dict(zip(cols, types)), 
                      engine=self._format._options['engine'] or vulkn.engines.Memory())

        if self._format._options['infer_schema']:
            sample_engine = self._format._options['sample_engine'] or vulkn.engines.Memory()
            _create_table(tmp_db, f'tmp_{table}', dict(zip(cols, types)), sample_engine)
            # TODO: Next release. Write without cat
            self._format.read(PosixStorage(uri), tmp_db, f'tmp_{table}')
            convert_dml = marshal_columns(tmp_db, f'tmp_{table}', database, table, cols, types)
            if ctx._conn.execute(convert_dml) != 0:
                raise Exception('Unable to load data')
            ctx.dropTable(tmp_db, f'tmp_{table}')
        else:
            # TODO: Next release. Write without cat
            self._format.read(PosixStorage(uri), database, table)

        return BaseTableDataTable(ctx, database, table).select('*')
        

def snake_case(name):
    f = ['#','/','<','(',')','%','|','@','.','-',' ','[',']','!','%','+','__']
    r = ['num','_','lt','','','','','','_','_','_','_','_','','pct','_','_']
    s = name.replace('dB', 'db').replace('MHz', 'mhz').replace('Mbit/s', 'mbps')
    s = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', s)
    s = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s)
    for fs, rs in dict(zip(f,r)).items():
        s = s.replace(fs, rs)
    return s.lower()
    

def infer_column_types(database, table, columns, sample=0, allow_enums=True):
    tests = []
    results = []
    ctx = vulkn.Vulkn()
    for col in columns:
        r = ctx._conn.select(infer_column_type(database, table, col, sample)).to_records()
        if len(r) > 0:
            tests.append(r[0])
    for i, col in enumerate(tests):
        r = ''
        if tests[i]['recommended_type'] == 'String' and tests[i]['enum'] and allow_enums:
            enum = ["'{}'={}".format(st, str(idx+1))
                for idx, st in enumerate(sorted(ast.literal_eval(tests[i]['sample'])))]
            r = "Enum({})".format(', '.join(enum))
        else:
            r = tests[i]['recommended_type']
        if tests[i]['nullable']:
            r = 'Nullable({})'.format(r)
        if tests[i]['low_cardinality']:
            r = 'LowCardinality({})'.format(r)
        results.append(r)
    return results


def marshal_columns(src_database, src_table, target_database, target_table, columns, types):
    cols = []
    for i, t in enumerate(types):
        c = 'nullIf(trim(toString("{}")), \'\')'.format(columns[i])
        if t in ('DateTime', 'Date'):
            c = 'parseDateTimeBestEffortOrNull({})'.format(c)
        cols.append(c)
    columns = ','.join(cols)
    database = target_database
    table = target_table
    sql = f'INSERT INTO {database}.{table} SELECT {columns} FROM {src_database}.{src_table}'
    return sqlformat(sql)


def infer_column_type(database, table, column, sample=0):
    sample = 0
    sql = """
        SELECT
            '{column}' AS col_name,
            any(nullable) AS nullable,
            (uniq_approx < 6 AND recommended_type = 'String') AS enum,
            (uniq_approx > 10 AND uniq_approx < 1000000 AND recommended_type = 'String') AS low_cardinality,
            recommended_type,
            any(sample) AS sample,
            any(uniq_approx) AS uniq_approx
        FROM (
            SELECT
                (null_cnt > 0) AS nullable,
                multiIf(
                    uint_cnt = non_null_cnt,
                        multiIf(
                            length(toString(toUInt8OrNull(col))) = length(col), 'UInt8',
                            length(toString(toUInt16OrNull(col))) = length(col), 'UInt16',
                            length(toString(toUInt32OrNull(col))) = length(col), 'UInt32',
                            length(toString(toUInt64OrNull(col))) = length(col), 'UInt64',
                            'String'),
                    int_cnt = non_null_cnt,
                        multiIf(
                            length(toString(toInt8OrNull(col))) = length(col), 'Int8',
                            length(toString(toInt16OrNull(col))) = length(col), 'Int16',
                            length(toString(toInt32OrNull(col))) = length(col), 'Int32',
                            length(toString(toInt64OrNull(col))) = length(col), 'Int64',
                            'String'),
                    float_cnt = non_null_cnt,
                        multiIf(
                            length(toString(toFloat32OrNull(col))) = length(col), 'Float32',
                            length(toString(toFloat64OrNull(col))) = length(col), 'Float64',
                            'String'),
                    datetime_cnt = non_null_cnt,
                        if(parseDateTimeBestEffortOrNull(col) IS NOT NULL, 'DateTime', 'String'),
                    date_cnt = non_null_cnt,
                        if(parseDateTimeBestEffortOrNull(col) IS NOT NULL, 'Date', 'String'),
                    'String') AS recommended_type,
                sample,
                if({sample} == 0, any(uniq_values), any(toUInt64(uniq_values*(100/{sample})))) AS uniq_approx
            FROM (
                SELECT
                    '{column}' AS col_name,
                    toString("{column}") AS col
                FROM {database}.{table_name}
                WHERE if({sample} == 0, 1, rand(1)%100 <= ({sample} - 1))
            ) ALL INNER JOIN (
                SELECT
                    col_name,
                    count() AS col_cnt,
                    countIf(col IS NULL) AS null_cnt,
                    col_cnt - null_cnt AS non_null_cnt,
                    sum(match(col, '^([\-0-9\.]*)$')) AS float_cnt,
                    sum(match(col, '^([\-0-9]*)$')) AS int_cnt,
                    sum(match(col, '^([0-9]*)$')) AS uint_cnt,
                    sum(match(col, '.*([0-9]{{2}}:[0-9]{{2}}).*')) AS datetime_cnt,
                    sum(match(col, '.*([0-9]{{2}}[/\-][0-9]{{2}}[/\-][0-9]{{2}}).*')) AS date_cnt,
                    uniqExact(col) AS uniq_values,
                    topK(10)(col) AS sample
                FROM (
                    SELECT
                        '{column}' AS col_name,
                        nullIf(trim(toString("{column}")), '') AS col
                    FROM {database}.{table_name}
                    WHERE if({sample} == 0, 1, rand(1)%100 <= ({sample} - 1))
                ) GROUP BY col_name
            ) USING (col_name)
            GROUP BY nullable, recommended_type, sample
        ) GROUP BY recommended_type, col_name
        ORDER BY
            recommended_type = 'UInt64' DESC
            , recommended_type = 'UInt32' DESC
            , recommended_type = 'UInt16' DESC
            , recommended_type = 'UInt8' DESC
            , recommended_type = 'Int64' DESC
            , recommended_type = 'Int32' DESC
            , recommended_type = 'Int16' DESC
            , recommended_type = 'Int8' DESC
            , recommended_type = 'Float64' DESC
            , recommended_type = 'Float32' DESC
            , recommended_type = 'DateTime' DESC
            , recommended_type = 'Date' DESC
            , recommended_type = 'String' DESC
        LIMIT 1
    """
    return sqlformat(sql.format(database=database, column=column, table_name=table, sample=sample))



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.types.string import String
from vulkn.types.url import URL
from vulkn.types.array import Array
from vulkn.types.base import TypeBase, Literal
from vulkn.types.numeric import Numeric
from vulkn.types.integer import UInt8, UInt16, UInt32, UInt64, Int8, Int16, Int32, Int64
from vulkn.types.float import Float, Float32, Float64
from vulkn.types.boolean import Boolean
from vulkn.types.date import Date, DateTime
from vulkn.types.url import URL
from vulkn.types.json import JSON
from vulkn.types.ipaddress import IP
from vulkn.types.array_vector import ArrayVector
from vulkn.types.column_vector import ColumnVector
from vulkn.types.merge_vector import MergeVector
from vulkn.types.column_type import ColumnType
from vulkn.types.uuid import UUID

Vector = ArrayVector

vS = String
vU = URL
vA = Array
vN = Numeric
vU8 = UInt8
vU16 = UInt16
vU32 = UInt32
vU64 = UInt64
vI8 = Int8
vI16 = Int16
vI32 = Int32
vI64 = Int64
vF32 = Float32
vF64 = Float64
vD = Date
vDT = DateTime
vDT32 = DateTime
col = lambda x: TypeBase(Literal(x))
c = col
null = c('NULL')
NULL = null


"""
TODO: Next release.

Strings:
- Add String function format
- Add searching strings https://clickhouse.yandex/docs/en/query_language/functions/string_search_functions/
- Add replacing strings https://clickhouse.yandex/docs/en/query_language/functions/string_replace_functions/
Type Conversion:
- https://clickhouse.yandex/docs/en/query_language/functions/type_conversion_functions/
Dates:
- Add toRelative*, add*, subtract* functions
Arrays
- https://clickhouse.yandex/docs/en/query_language/functions/array_functions/
Hash
- https://clickhouse.yandex/docs/en/query_language/functions/hash_functions/
Dictionaries
- https://clickhouse.yandex/docs/en/query_language/functions/ext_dict_functions/
ML
- https://clickhouse.yandex/docs/en/query_language/functions/machine_learning_functions/
Other
- https://clickhouse.yandex/docs/en/query_language/functions/other_functions/
"""



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from collections.abc import Mapping


class VulknDataStore(Mapping):
    def __init__(self, ctx, root=None):
        self._ctx = ctx
        self._root = root

    def __getattr__(self, attrname):
        if not self._root:
            sql = "SELECT name FROM system.databases WHERE name = {catalog:String}"
            t = self._ctx.q(sql).exec(catalog=attrname).to_records()
            if len(t) > 0:
                return VulknDataStore(ctx=self._ctx, root=t[0]['name'])
        else:
            sql = "SELECT name FROM system.tables WHERE database = {catalog:String} AND name = {table:String}"
            t = self._ctx.q(sql).exec(catalog=self._root, table=attrname).to_records()
            if len(t) > 0:
                return self._ctx.table(f'{self._root}.{attrname}')
            raise AttributeError

    def __getitem__(self, key):
        sql = "SELECT database, name FROM system.tables WHERE (database || '.' || name) = {key:String}"
        t = self._ctx.q(sql).exec(key=key).to_records()
        if len(t) > 0:
            return self._ctx.table(t[0]['database'], t[0]['name'])
        raise AttributeError

    def __iter__(self):
        sql = "SELECT database, name FROM system.tables ORDER BY database, name"
        d = {}
        for r in self._ctx.q(sql).exec().to_records():
            database = r['database']
            table = r['name']
            d[f'{database}.{table}'] = self._ctx.table(database, table)
        return iter(d)

    def __len__(self):
        sql = "SELECT count() AS count FROM system.tables"
        t = int(self._ctx.q(sql).exec().to_records()[0]['count'])
        return t

    def help(self):
        sql = "SELECT name FROM system.tables WHERE database = {catalog:String} ORDER BY name"
        return [n['name'] for n in self._ctx.q(sql).exec(catalog=self._catalog).to_records()]




# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


class ListOption:
    def __init__(self, name, *options):
        self.name = name
        self.options = options

    def __str__(self):
        r = ''
        if len(self.options) > 0:
            parameters = ', '.join(list(map(str, self.options)))
            r = '{keyword} ({parameters})'.format(keyword=self.name, parameters=parameters)
        return r


class KeyValueOption:
    def __init__(self, name, **options):
        self.name = name
        self.options = options

    def __str__(self):
        r = ''
        if len(self.options) > 0:
            parameters = ', '.join(
                ['{option}={value}'.format(option=k,value=v) for k, v in self.options.items()])
            r = '{keyword} {parameters}'.format(keyword=self.name, parameters=parameters)
        return r


class Simple:
    storage_type = '' 

    def __init__(self, settings=None):
        self.settings = KeyValueOption('SETTINGS', **(settings or {}))

    def __str__(self):
        r = '{}()'.format(self.storage_type)
        r += ' {} '.format(self.settings)
        return r.strip()


class Memory(Simple):
    storage_type = 'Memory'


class Log(Simple):
    storage_type = 'Log'


class TinyLog(Simple):
    storage_type = 'TinyLog'


class StripeLog(Simple):
    storage_type = 'StripeLog'


class Set(Simple):
    storage_type = 'Set'


class GenericMergeTree:
    storage_type = 'MergeTree'

    def __init__(self, partition_by=None, order_by=None, sample_by=None, settings=None):
        self.partition_by = ListOption('PARTITION BY', *((partition_by) or ()))
        self.order_by = ListOption('ORDER BY', *((order_by) or ('tuple()',)))
        self.sample_by = ListOption('SAMPLE BY', *((sample_by) or ()))
        self.settings = KeyValueOption('SETTINGS', **(settings or {}))


class MergeTree(GenericMergeTree):
    storage_type = 'MergeTree'

    def __init__(self,
                 partition_by=None,
                 order_by=None,
                 primary_key=None,
                 sample_by=None,
                 ttl=None,
                 settings=None):
        self.primary_key = ListOption('PRIMARY KEY', *((primary_key) or ()))
        self.ttl = ListOption('TTL', *((ttl) or ()))
        super().__init__(
            partition_by=partition_by, order_by=order_by, sample_by=sample_by, settings=settings)

    def __str__(self):
        r = '{}()'.format(self.storage_type)
        r = (r + ' {}'.format(self.partition_by)).strip()
        r = (r + ' {}'.format(self.order_by)).strip()
        r = (r + ' {}'.format(self.primary_key)).strip()
        r = (r + ' {}'.format(self.sample_by)).strip()
        r = (r + ' {}'.format(self.ttl)).strip()
        r = (r + ' {}'.format(self.settings)).strip() 
        return r.strip()


class ReplacingMergeTree(GenericMergeTree):
    storage_type = 'ReplacingMergeTree'

    def __init__(self, 
                 partition_by=None,
                 order_by=None,
                 sample_by=None,
                 version=None,
                 settings=None):
        self.version = version
        super().__init__(
            partition_by=partition_by, order_by=order_by, sample_by=sample_by, settings=settings)

    def __str__(self):
        r = self.storage_type
        r += '({})'.format(self.version) if self.version else '()'
        r = (r + ' {}'.format(self.partition_by)).strip()
        r = (r + ' {}'.format(self.order_by)).strip()
        r = (r + ' {}'.format(self.sample_by)).strip()
        r = (r + ' {}'.format(self.settings)).strip()
        return r.strip()


class SummingMergeTree(GenericMergeTree):
    storage_type = 'SummingMergeTree'

    def __init__(self,
                 partition_by=None,
                 order_by=None,
                 sample_by=None,
                 columns=None,
                 settings=None):
        self.columns = columns
        super().__init__(
            partition_by=partition_by, order_by=order_by, sample_by=sample_by, settings=settings)

    def __str__(self):
        r = self.storage_type
        r += '({})'.format(self.columns) if self.columns else '()'
        r = (r + ' {}'.format(self.partition_by)).strip()
        r = (r + ' {}'.format(self.order_by)).strip()
        r = (r + ' {}'.format(self.sample_by)).strip()
        r = (r + ' {}'.format(self.settings)).strip()
        return r.strip()


class AggregatingMergeTree(GenericMergeTree):
    storage_type = 'AggregatingMergeTree'

    def __str__(self):
        r = self.storage_type
        r = (r + ' {}'.format(self.partition_by)).strip()
        r = (r + ' {}'.format(self.order_by)).strip()
        r = (r + ' {}'.format(self.sample_by)).strip()
        r = (r + ' {}'.format(self.settings)).strip()
        return r.strip()


class CollapsingMergeTree(GenericMergeTree):
    storage_type = 'CollapsingMergeTree'
    
    def __init__(self, partition_by=None, order_by=None, sample_by=None, sign=None, settings=None): 
        self.sign = sign
        super().__init__(
            partition_by=partition_by, order_by=order_by, sample_by=sample_by, settings=settings)

    def __str__(self):
        r = self.storage_type
        r += '({})'.format(self.sign) if self.sign else '()'
        r = (r + ' {}'.format(self.partition_by)).strip()
        r = (r + ' {}'.format(self.order_by)).strip()
        r = (r + ' {}'.format(self.sample_by)).strip()
        r = (r + ' {}'.format(self.settings)).strip()
        return r.strip()


class Join:
    def __init__(self, join_strictness, join_type, keys):
        self.join_strictness = join_strictness
        self.join_type = join_type
        self.keys = keys

    def __str__(self):
        r = 'Join({strictness}, {type}, {keys})'.format(
            strictness=self.join_strictness, type=self.join_type, keys=','.join(self.keys))
        return r
        

class BufferProfile:
    def __init__(self,
                 num_layers=16,
                 min_time=10,
                 max_time=100,
                 min_rows=10000,
                 max_rows=1000000,
                 min_bytes=10000000,
                 max_bytes=100000000,
                 prefix=None,
                 postfix=None):
        self.num_layers = num_layers
        self.min_time = min_time
        self.max_time = max_time
        self.min_rows = min_rows
        self.max_rows = max_rows
        self.min_bytes = min_bytes
        self.max_bytes = max_bytes
        self.prefix = prefix

    def DDL(self, database, table, prefix=None, postfix=None):
        buffer_table = table
        if prefix:
            buffer_table = '{}_{}'.format(prefix, buffer_table)
        if postfix:
            buffer_table = '{}_{}'.format(buffer_table, postfix)
        buffer_table_ddl = f'CREATE TABLE {database}.{buffer_table} AS {database}.{table} ENGINE = '
        buffer_profile = self.profile(database, table)
        return f'{buffer_table_ddl} {buffer_profile}'
    
    def profile(self, database, table):
        profile_options = ','.join(list(map(str, [database,
                                                  table,
                                                  self.num_layers,
                                                  self.min_time,
                                                  self.max_time,
                                                  self.min_rows,
                                                  self.max_rows,
                                                  self.min_bytes,
                                                  self.max_bytes])))
        return f'Buffer({profile_options})'



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import collections
import csv
import ast
import itertools
import logging
import ciso8601
import pandas as pd 
import numpy as np
import io
import datetime
import tabulate
from typing import List


from vulkn.utils import timer


log = logging.getLogger()
np.set_printoptions(suppress=True)
tabulate.PRESERVE_WHITESPACE = True


class TypeInfoMap:
    CONVERSION_MAP = {
        'UInt8': { 'native': lambda x: int(x), 'pandas': np.dtype('u1') },
        'UInt16': { 'native': lambda x: int(x), 'pandas': np.dtype('u2') },
        'UInt32': { 'native': lambda x: int(x), 'pandas': np.dtype('u4') },
        'UInt64': { 'native': lambda x: int(x), 'pandas': np.dtype('u8') },
        'Int8': { 'native': lambda x: int(x), 'pandas': np.dtype('i1') },
        'Int16': { 'native': lambda x: int(x), 'pandas': np.dtype('i2') },
        'Int32': { 'native': lambda x: int(x), 'pandas': np.dtype('i4') },
        'Int64': { 'native': lambda x: int(x), 'pandas': np.dtype('i8') },
        'Float32': { 'native': lambda x: float(x), 'pandas': np.dtype('f4') },
        'Float64': { 'native': lambda x: float(x), 'pandas': np.dtype('f8') },
        'Decimal': { 'native': lambda x: float(x), 'pandas': np.float },
        'DateTime': { 'native': lambda x: ciso8601.parse_datetime(x), 'pandas': np.dtype('M') },
        'Date': { 'native': lambda x: ciso8601.parse_datetime(x), 'pandas': np.dtype('M') }
    }

    NATIVE_CONVERSION_MAP = dict((k, v['native']) for k, v in CONVERSION_MAP.items())
    PANDAS_CONVERSION_MAP = dict((k, v['pandas']) for k, v in CONVERSION_MAP.items())

    @staticmethod
    def native_marshall_map(chtypes):
        types = []
        for t in chtypes:
            if t in TypeInfoMap.NATIVE_CONVERSION_MAP:
                types.append(TypeInfoMap.NATIVE_CONVERSION_MAP[t])
            else:
                if t.startswith('Array(') or t.startswith('Tuple('):
                    types.append(ast.literal_eval)
                else:
                    types.append(str)
        return types

    @staticmethod
    def pandas_marshall_map(columns, chtypes):
        # TODO: Fix type conversion of nested types or adopt Arrow/Parquet approach.
        types = {}
        for c, t in zip(columns, chtypes):
            if t in TypeInfoMap.PANDAS_CONVERSION_MAP:
                types[c] = TypeInfoMap.PANDAS_CONVERSION_MAP[t]
        return types


def ConsoleOutputFormatter(results: List[List[str]], columns: List[str]) -> str:
    return '\n{results}\n\n({rowcount} row{plural})\n'.format(
                                    results=tabulate.tabulate(results, columns, tablefmt='simple'),
                                    rowcount=len(results),
                                    plural='' if len(results) == 1 else 's')


class Row(collections.OrderedDict):
    def __init__(self, column_names, types, data):
        d = [(column_names[i], None if r == '\\N' else types[i](r)) for i, r in enumerate(data)]
        super(Row, self).__init__(d)


class RecordSet():
    def __init__(self, raw):
        # TODO: Raw is a cached version of the result as a string/blob that is re-interpreted 
        # each time. This makes no sense. Modify interface to use clickhouse-driver, Parquet -> Arrow
        # or another method under the hood.
        self._data = raw
        self._columns = None
        self._chtypes = None
        self._pytypes = None
        self._pdtypes = None
        self._marshall_info_ = None
        self._result = None

    @property
    def columns(self):
        if not self._columns:
            c = pd.read_csv(io.StringIO(self._data), sep='\t', header=None, na_values='\\N', nrows=1)
            self._columns = c.values.tolist()[0]
        return self._columns

    @property
    def chtypes(self):
        if not self._chtypes:
            c = pd.read_csv(io.StringIO(self._data),
                            skiprows=([0]),
                            sep='\t',
                            header=None,
                            na_values='\\N',
                            nrows=1)
            self._chtypes = c.values.tolist()[0]
        return self._chtypes

    @property
    def pytypes(self):
        # TODO: Unify with TypeMap
        if not self._pytypes:
            types = []
            for t in self.chtypes:
                if t.startswith('UInt') or t.startswith('Int'):
                    types.append(int)
                elif t.startswith('Float') or t.startswith('Decimal'): 
                    types.append(float)
                elif t == 'DateTime':
                    types.append(datetime.datetime)
                elif t == 'Date':
                    types.append(datetime.date)
                elif t.startswith('Array(') or t.startswith('Tuple('):
                    types.append(object)
                else:
                    types.append(str)
            self._pytypes = types
        return self._pytypes

    @property
    def pdtypes(self):
        if not self._pdtypes:
            self._pdtypes = TypeInfoMap.pandas_marshall_map(self.columns, self.chtypes)
        return self._pdtypes

    @property
    def _marshall_info(self):
        if not self._marshall_info_:
            self._marshall_info_ = TypeInfoMap.native_marshall_map(self.chtypes)
        return self._marshall_info_

    def to_pandas(self):
        if not self._result:
            self._result = pd.read_csv(
                io.StringIO(self._data),
                skiprows=([1]), sep='\t',
                header=0,
                na_values='\\N',
                parse_dates=True,
                engine='c',
                infer_datetime_format=True)
        return self._result

    def to_dict(self):
        return self.to_pandas().to_dict('record')

    def to_records(self):
        return self.to_pandas().to_dict('record')
 
    def to_list(self):
        return self.to_recarray().tolist()

    def to_recarray(self):        
        return self.to_pandas().to_records(index=False)

    def show(self):        
        table = self.to_pandas().values[0:1000].tolist()
        table = [[i+1]+r for i, r in enumerate(table)]
        print(ConsoleOutputFormatter(table, ['row'] + self.columns))

    def show_pandas(self, num_rows=50):
        pd.options.display.max_rows = num_rows
        return self.to_pandas()

    s = property(show)
    r = property(to_records)
    p = property(to_pandas)
    d = property(to_dict)
    l = property(to_list)
    n = property(to_recarray)

    def __len__(self):
        return len(self.to_pandas())

    def __getitem__(self, key):
        cnt = itertools.count(1)
        if isinstance(key, slice):
            for row in csv.reader(self._data[key.start + 2:key.stop], delimiter='\t', quotechar='"'):
                c = next(cnt)
                if(c%100000 == 0):
                    log.debug(f"Read {c} records...")
                yield Row(self.columns, self._marshal_info, row)
            c = next(cnt)
            log.debug(f"Read {c} records...")
        elif isinstance(key, int):
            for row in csv.reader([self._data[key + 2]], delimiter='\t', quotechar='"'):
                yield Row(self.columns, self._marshal_info, row)




# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


# TODO: Next release.

class Tuple:
    def __init__(self, *args, keys=True):
        self._column = None
        self._tuple = args
        self._use_keys = keys

    def __getattr__(self, key):
        try:
            return self._tuple[key]
        except KeyError:
            raise AttributeError

    @classmethod
    def c(self, column, keys):
        t = Tuple()
        t._column = column
        t._use_keys = True
        t._keys = keys
        structure
        return t


"""
k = Tuple(['a','b']).alias('z')
(k['a'] / k['d'])

df.select(
    Tuple.c('q', structure=[('a','b'),('c','d')]))
"""


# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import subprocess
import os


def sqlformat(query, oneline=True, hilite=False):
    env = {'LC_ALL': 'C'}
    cmd = ['clickhouse-format']
    if oneline:
        cmd += ['--oneline']
    if hilite:
        cmd += ['--hilite']
    try:
        p = subprocess.run(cmd,
                           stdout=subprocess.PIPE,
                           input=query.strip().replace('\\', ''),
                           env=env,
                           encoding='ascii')
        if p.returncode != 0:
            raise Exception
    except:
        print(query.strip().replace('\\', ''))
        raise
    return str('{};'.format(p.stdout.strip()))


def net_list_tables(path):
    return os.listdir(path)


def net_list_parts(path, table):
    return [
        d for d in os.listdir('{path}/{table}'.format(path=path, table=table)) 
            if os.path.isdir('{path}/{table}/{d}'.format(path=path, table=table, d=d))
            and d != 'detached']


def net_list_detached(path, table):
    return os.listdir('{path}/{table}/detached'.format(path=path, table=table))


def net_show_schema(path, table):
    with open(f'{path}/{table}/{table}.sql', 'r') as f:
        return '{};'.format(''.join(f.readlines()).strip())
           

def net_attach(path, table):
    schema = net_show_schema(path, table)
    # create symlink
    # run query
    # execute schema query above


def net_archive(path, table):
    pass
    # detach table
    # copy data to fs
    # attach table
    # copy metadata to archive
    # drop table



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import string
import socket
from timeit import default_timer
from functools import wraps


class LogLevels:
    SQL = 11


def get_next_free_socket(host, ports):
    for port in ports:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = sock.connect_ex((host, port))
        sock.close()
        if result != 0:
            return port
    return 0


def timer(f):
    @wraps(f)
    def wrap(*args, **kwargs):
        import vulkn.session
        if vulkn.session.timing:
            start = default_timer()
        r = f(*args, **kwargs)
        if vulkn.session.timing:
            # TODO: Overhead on avg PC is ~ 40ms for forking and running CLI vs direct call. 
            # Need to remove this overhead with HTTP, hybrid or persistent client.
            # Timing function roughly takes this into consideration.
            # 40ms overhead is a reflection on Vulkn implementation, not ClickHouse.
            overhead = 0.040
            print('Time elapsed: {:0.3f} ms'.format(abs(default_timer() - start - overhead)*1000))
        return r
    return wrap


def item_index(index):
    start = 1
    stop = None
    if isinstance(index, slice):
        stop = 1 if index.stop == 0 else index.stop
        start = index.start
    else:
        start = index
    return (start, stop)


def column_list(columns):
    from vulkn.types import ArrayVector, ColumnVector
    rangeCols = []
    selectExprList = []
    for idx, column in enumerate(columns):
        if isinstance(column, ArrayVector):
            rangeCols.append(str(column))
        if isinstance(column, ColumnVector):
            if hasattr(column, 'vector_name'):
                selectExprList.append('c{index}.v AS {name}'.format(index=str(idx),
                                                                    name=column.vector_name))
            else:    
                selectExprList.append('c{index}.v AS col{index}'.format(index=str(idx)))
            subquery = 'SELECT i, v FROM {column}'.format(str(column))
            if idx == 0:
                rangeCols.append('FROM ({}) c0'.format(subquery))
            else:
                rangeCols.append(
                    'INNER JOIN ({subquery}) c{index} ON (c{prevIndex}.i = c{index}.i)'.format(
                        subquery=subquery, index=str(idx), prevIndex=str(i-1)))
    q = "SELECT {selectExpr} {rangeCols}".format(selectExpr=', '.join(selectExprList), 
                                                 rangeCols=' '.join(rangeCols))
    return q


class singleton:
    def __init__(self, cls):
        self.cls = cls
        self.instance = None
    
    def __call__(self, *args, **kwargs):
        if self.instance == None:
            self.instance = self.cls(*args, **kwargs)
        return self.instance


class VulknSQLFormatter(string.Formatter):
    # TODO: Next release. Unify with RecordSet conversion map
    CONVERSION_MAP = {
        'String': lambda x: "'{}'".format(str(x).encode('unicode-escape').decode()),
        'UInt8': lambda x: int(x),
        'UInt16': lambda x: int(x),
        'UInt32': lambda x: int(x),
        'UInt64': lambda x: int(x),
        'Int8': lambda x: int(x),
        'Int16': lambda x: int(x),
        'Int32': lambda x: int(x),
        'Int64': lambda x: int(x),
        'Float32': lambda x: float(x),
        'Float64': lambda x: float(x),        
        'DateTime': lambda x: "'{}'".format(str(x).encode('unicode-escape').decode()),
        'Date': lambda x: "'{}'".format(str(x).encode('unicode-escape').decode())
    }

    def format_field(self, value, format_spec):
        if format_spec in self.CONVERSION_MAP:
            return '{}'.format(self.CONVERSION_MAP[format_spec](value))
        else:
            return super().format_field(value, format_spec)



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


def quote_literal(value: any) -> any:
    if isinstance(value, str):
        return "'{}'".format(value)
    elif is_literal_type(value):
        return value
    else:
        if isinstance(value, list):
            a = []
            for v in value:
                if isinstance(v, str):
                    a.append("'{}'".format(str(v).encode('unicode-escape').decode()))
                else:
                    a.append(str(v))
            return '[{}]'.format(', '.join(a))
            #return '[{}]'.format(','.join(["'{}'".format(str(v).encode('unicode-escape').decode()) for v in value]))
            #return '[{}]'.format(','.join(["{}".format(str(v).encode('unicode-escape').decode()) for v in value]))
        else:
            return str(value)


format_python_type = quote_literal


class ColumnBaseMixIn:
    def alias(self, alias: str) -> any:
        r = type(self)(FunctionExpression('alias', str(self), alias))
        return r

    as_ = alias

    def cast(self, type_arg: str) -> any:
        from vulkn.types.string import String
        from vulkn.types.array import Array
        from vulkn.types.boolean import Boolean
        from vulkn.types.integer import Int8, Int16, Int32, Int64
        from vulkn.types.integer import UInt8, UInt16, UInt32, UInt64
        from vulkn.types.date import Date, DateTime
                
        k = getattr(type_arg, 'CAST', type_arg)
        return locals()[k](FunctionExpression('cast', Literal(str(self)), k))


class Literal(ColumnBaseMixIn):
    def __init__(self, value: any) -> None:
        self._value = value

    def __str__(self) -> str:
        if self._value is None:
            return 'NULL'
        else:
            return str(self._value)


Expression = Literal


class FunctionExpression:
    def __init__(self, el: str, *args: list) -> None:
        self._el = el
        self._args = args

    def __str__(self) -> str:
        if self._el == 'alias':
            return '{} AS `{}`'.format(str(self._args[0]), self._args[1])
        else:
            args = ["'{}'".format(a) if isinstance(a, str) else str(a) for a in self._args]
            return '{}({})'.format(self._el, ','.join(map(str, args)))


func = FunctionExpression


def is_literal_type(t) -> bool:
    return (isinstance(t, Literal) or 
            isinstance(t, Expression) or 
            isinstance(t, FunctionExpression))


class TypeBase(ColumnBaseMixIn):
    _METHODS = {}
    TYPE = 'String'
    CAST = TYPE

    def __init__(self, value: any=None, name: str=None, n: str=None) -> None:
        self._value = Literal('"{}"'.format(name or n)) if (name or n) else value

    def __str__(self) -> str:
        return (str(self._value) if is_literal_type(self._value) 
                else format_python_type(self._value))

    @property
    def sql(self):
        return str(self)

    @property
    def methods(self) -> dict:
        r = {}
        if 'methods' in dir(super()):
            r.update(super().methods)
        r.update(self._METHODS)
        return r

    def __getattr__(self, name: str) -> any:
        def _method():
            from vulkn.types.string import String
            from vulkn.types.array import Array
            from vulkn.types.boolean import Boolean
            from vulkn.types.integer import Int8, Int16, Int32, Int64
            from vulkn.types.integer import UInt8, UInt16, UInt32, UInt64
            from vulkn.types.date import Date, DateTime

            rtype = (locals()[self.methods[name]] 
                if self.methods[name] is not None
                else type(self))
            
            return rtype(FunctionExpression(name, self._value))

        if name in self.methods:
            return _method
        else:
            raise AttributeError
 
    def _in(self, operator, right):
        from vulkn.types.integer import UInt8
        from vulkn.datatable import VulknDataTable
        in_arg = None
        if hasattr(right, 'show_sql'):
            in_arg = right.show_sql()[0:-1]
        elif isinstance(right, list):
            if isinstance(right[0], str):
                in_arg = "'{}'".format("','".join(map(str, right)))
            else:
                in_arg = ','.join(map(str, right))
        else:
            in_arg = str(right)
        return UInt8(Literal('{} {} ({})'.format(self._value, operator, in_arg)))

    def In(self, right):
        return self._in('IN', right)

    in_ = In

    def notIn(self, right):
        return self._in('NOT IN', right)

    not_in = notIn
    not_in_ = notIn

    def globalIn(self, right):
        return self._in('GLOBAL IN', right)

    global_in = globalIn
    global_in_ = globalIn

    def globalNotIn(self, right):
        return self._in('GLOBAL NOT IN', right)

    global_not_in = globalNotIn
    global_not_in_ = globalNotIn

    def isNull(self):
        from vulkn.types.integer import UInt8
        return UInt8(func('isNull', self._value))

    is_null = isNull
    isn = isNull

    def isNotNull(self):
        from vulkn.types.integer import UInt8
        return UInt8(func('isNotNull', self._value))

    is_not_null = isNotNull
    isnn = isNotNull

    def ifNull(self, other):
        return Literal(func('ifNull', self._value, other))

    if_null = ifNull

    def nullIf(self, test):
        return Literal(func('nullIf', self._value, test))

    null_if = nullIf

    def assumeNotNull(self):
        return type(self)(func('assumeNotNull', self._value))

    assume_not_null = assumeNotNull
    nn = assumeNotNull

    def toNullable(self):
        return type(self)(func('toNullable', self._value))

    to_nullable = toNullable
    tn = toNullable

    def between(self, start, end):
        from vulkn.types.integer import UInt8
        return UInt8(Literal('{} BETWEEN {} AND {}'.format(self._value, start, end)))

    bt = between

    def notBetween(self, start, end):
        from vulkn.types.integer import UInt8
        return UInt8(Literal('{} NOT BETWEEN {} AND {}'.format(self._value, start, end)))

    not_between = notBetween
    nbt = notBetween

    def typename(self):
        from vulkn.types.string import String
        return String(Literal(f'toTypeName({self._value})'))

    def __eq__(self, right) -> any:
        from vulkn.types.integer import UInt8
        return UInt8(func('equals', self._value, right))

    eq = __eq__
    equals = __eq__
        
    def __ne__(self, right) -> any:
        from vulkn.types.integer import UInt8
        return UInt8(func('notEquals', self._value, right))

    ne = __ne__
    notEquals = __ne__
    not_equals = __ne__

    def __gt__(self, right) -> any:
        from vulkn.types.integer import UInt8
        return UInt8(func('greater', self._value, right))

    gt = __gt__
    greater = __gt__
        
    def __lt__(self, right) -> any:
        from vulkn.types.integer import UInt8
        return UInt8(func('less', self._value, right))

    lt = __lt__
    less = __lt__

    def __ge__(self, right) -> any:
        from vulkn.types.integer import UInt8
        return UInt8(func('greaterOrEquals', self._value, right))
        
    ge = __ge__
    greaterOrEquals = __ge__
    greater_or_equals = __ge__

    def __le__(self, right) -> any:
        from vulkn.types.integer import UInt8
        return UInt8(func('lessOrEquals', self._value, right))

    le = __le__
    lessOrEquals = __le__
    less_or_equals = __le__

    def and_(self, right) -> any:
        from vulkn.types.integer import UInt8
        return UInt8(func('and', self._value, right))

    def __and__(self, right) -> any:
        from vulkn.types.integer import UInt8
        return UInt8(func('bitAnd', self._value, right))

    bitAnd = __and__
    bit_and = __and__

    def or_(self, right) -> any:
        from vulkn.types.integer import UInt8
        return UInt8(func('or', self._value, right))

    def __or__(self, right) -> any:
        from vulkn.types.integer import UInt8
        return UInt8(func('bitOr', self._value, right))

    bitOr = __or__
    bit_or = __or__
        
    def not_(self) -> any:
        from vulkn.types.integer import UInt8
        return UInt8(func('not', self._value))

    def __invert__(self) -> any:
        from vulkn.types.integer import UInt8
        return UInt8(func('bitNot', self._value))

    bitNot = __invert__
    bit_not = __invert__

    def __xor__(self, right) -> any:
        from vulkn.types.integer import UInt8
        return UInt8(func('xor', self._value, right))

    xor = __xor__
    xor_ = __xor__
    bitXor = __xor__
    bit_xor = __xor__



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from enum import Enum, unique


def nan_to_none(value):
    try:
        if pd.np.isnan(value):
            return None
    except:
        pass
    return value


class ColumnType:
    @unique
    class DefaultKind(Enum):
        MATERIALIZED = 'MATERIALIZED'
        ALIAS = 'ALIAS'
        DEFAULT = 'DEFAULT'
        NONE = None

    def __init__(self, name, col_type, default_kind=None, default_expression=None, compression_codec=None):
        self._name = name
        self._col_type = col_type
        self._default_kind = nan_to_none(default_kind)
        self._default_expression = nan_to_none(default_expression)
        self._compression_codec = nan_to_none(compression_codec)
    
    def __repr__(self):
        _none_or_quoted = lambda value: 'None' if value is None else "'{}'".format(value)
        r = 'ColumnType({name}, {col_type}, default_kind={default_kind}, default_expression={default_expression} compression_codec={compression_codec})'
        return r.format(name="'{}'".format(self._name),
                        col_type="'{}'".format(self._col_type),
                        default_kind=_none_or_quoted(self._default_kind),
                        default_expression=_none_or_quoted(self._default_expression),
                        compression_codec=_none_or_quoted(self._compression_codec))

    def __str__(self):
        opts = ' '.join(filter(lambda x: x is not None,
                        [self._default_kind, self._default_expression, self._compression_codec]))
        return '{name} {col_type}{opts}'.format(name=self._name,
                                                col_type=self._col_type,
                                                opts=opts if len(opts) > 0 else '')



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


VulknConfig = {
    'metastore': '/home/kudos/vulkn/metastore',
    'tmp': 'vulkn_tmp'
}

VERSION_MAJOR = 19
VERSION_MINOR = 0
VERSION_PATCH = 6
VERSION_TEST = None
VERSION = '.'.join([str(VERSION_MAJOR), str(VERSION_MINOR), str(VERSION_PATCH)])

if VERSION_TEST:
    VERSION += '.' + VERSION_TEST

FLAGS = [
    'secure'
]

AUTH_OPTS = {
    'host': 'localhost',
    'port': 9000,
    'user': 'default',
    'password': None,
    'database': 'default'
}

EXTERNAL_FILE_OPTS = {
    'file': '-',
    'name': 'data',
    'format': 'TabSeparated',
    'structure': None,
    'types': None
}

SETTINGS = [
    'max_client_network_bandwidth',
    'server_logs_file',
    'min_compress_block_size',
    'max_compress_block_size',
    'max_block_size',
    'max_insert_block_size',
    'min_insert_block_size_rows',
    'min_insert_block_size_bytes',
    'max_threads',
    'max_read_buffer_size',
    'max_distributed_connections',
    'max_query_size',
    'interactive_delay',
    'connect_timeout',
    'connect_timeout_with_failover_ms',
    'receive_timeout',
    'send_timeout',
    'tcp_keep_alive_timeout',
    'queue_max_wait_ms',
    'poll_interval',
    'distributed_connections_pool_size',
    'connections_with_failover_max_tries',
    'use_uncompressed_cache',
    'replace_running_query',
    'background_pool_size',
    'background_schedule_pool_size',
    'distributed_directory_monitor_sleep_time_ms',
    'distributed_directory_monitor_batch_inserts',
    'optimize_move_to_prewhere',
    'replication_alter_partitions_sync',
    'replication_alter_columns_timeout',
    'load_balancing',
    'totals_mode',
    'totals_auto_threshold',
    'compile_expressions',
    'min_count_to_compile',
    'group_by_two_level_threshold',
    'group_by_two_level_threshold_bytes',
    'distributed_aggregation_memory_efficient',
    'aggregation_memory_efficient_merge_threads',
    'max_parallel_replicas',
    'parallel_replicas_count',
    'parallel_replica_offset',
    'skip_unavailable_shards',
    'distributed_group_by_no_merge',
    'merge_tree_min_rows_for_concurrent_read',
    'merge_tree_min_rows_for_seek',
    'merge_tree_coarse_index_granularity',
    'merge_tree_max_rows_to_use_cache',
    'merge_tree_uniform_read_distribution',
    'mysql_max_rows_to_insert',
    'optimize_min_equality_disjunction_chain_length',
    'min_bytes_to_use_direct_io',
    'force_index_by_date',
    'force_primary_key',
    'mark_cache_min_lifetime',
    'max_streams_to_max_threads_ratio',
    'network_compression_method',
    'network_zstd_compression_level',
    'log_queries',
    'log_queries_cut_to_length',
    'distributed_product_mode',
    'max_concurrent_queries_for_user',
    'insert_deduplicate',
    'insert_sample_with_metadata',
    'insert_quorum',
    'insert_quorum_timeout',
    'select_sequential_consistency',
    'table_function_remote_max_addresses',
    'read_backoff_min_latency_ms',
    'read_backoff_max_throughput',
    'read_backoff_min_interval_between_events_ms',
    'read_backoff_min_events',
    'memory_tracker_fault_probability',
    'enable_http_compression',
    'http_zlib_compression_level',
    'http_native_compression_disable_checksumming_on_decompress',
    'count_distinct_implementation',
    'output_format_write_statistics',
    'add_http_cors_header',
    'input_format_skip_unknown_fields',
    'input_format_import_nested_json',
    'input_format_values_interpret_expressions',
    'output_format_json_quote_64bit_integers',
    'output_format_json_quote_denormals',
    'output_format_json_escape_forward_slashes',
    'output_format_pretty_max_rows',
    'output_format_pretty_max_column_pad_width',
    'output_format_pretty_color',
    'use_client_time_zone',
    'send_progress_in_http_headers',
    'http_headers_progress_interval_ms',
    'fsync_metadata',
    'input_format_allow_errors_num',
    'input_format_allow_errors_ratio',
    'join_use_nulls',
    'join_default_strictness',
    'preferred_block_size_bytes',
    'max_replica_delay_for_distributed_queries',
    'fallback_to_stale_replicas_for_distributed_queries',
    'preferred_max_column_in_block_size_bytes',
    'insert_distributed_sync',
    'insert_distributed_timeout',
    'distributed_ddl_task_timeout',
    'stream_flush_interval_ms',
    'format_schema',
    'insert_allow_materialized_columns',
    'http_connection_timeout',
    'http_send_timeout',
    'http_receive_timeout',
    'optimize_throw_if_noop',
    'use_index_for_in_with_subqueries',
    'empty_result_for_aggregation_by_empty_set',
    'allow_distributed_ddl',
    'odbc_max_field_size',
    'max_rows_to_read',
    'max_bytes_to_read',
    'read_overflow_mode',
    'max_rows_to_group_by',
    'group_by_overflow_mode',
    'max_bytes_before_external_group_by',
    'max_rows_to_sort',
    'max_bytes_to_sort',
    'sort_overflow_mode',
    'max_bytes_before_external_sort',
    'max_bytes_before_remerge_sort',
    'max_result_rows',
    'max_result_bytes',
    'result_overflow_mode',
    'max_execution_time',
    'timeout_overflow_mode',
    'min_execution_speed',
    'timeout_before_checking_execution_speed',
    'max_columns_to_read',
    'max_temporary_columns',
    'max_temporary_non_const_columns',
    'max_subquery_depth',
    'max_pipeline_depth',
    'max_ast_depth',
    'max_ast_elements',
    'max_expanded_ast_elements',
    'max_rows_in_set',
    'max_bytes_in_set',
    'set_overflow_mode',
    'max_rows_in_join',
    'max_bytes_in_join',
    'join_overflow_mode',
    'max_rows_to_transfer',
    'max_bytes_to_transfer',
    'transfer_overflow_mode',
    'max_rows_in_distinct',
    'max_bytes_in_distinct',
    'distinct_overflow_mode',
    'max_memory_usage',
    'max_memory_usage_for_user',
    'max_memory_usage_for_all_queries',
    'max_network_bandwidth',
    'max_network_bytes',
    'max_network_bandwidth_for_user',
    'max_network_bandwidth_for_all_users',
    'format_csv_delimiter',
    'format_csv_allow_single_quotes',
    'format_csv_allow_double_quotes',
    'date_time_input_format',
    'log_profile_events',
    'log_query_settings',
    'log_query_threads',
    'send_logs_level',
    'enable_optimize_predicate_expression',
    'low_cardinality_max_dictionary_size',
    'low_cardinality_use_single_dictionary_for_part',
    'allow_experimental_low_cardinality_type',
    'decimal_check_overflow',
    'prefer_localhost_replica',
    'load_balancing',
    'max_fetch_partition_retries_count',
    'asterisk_left_columns_only',
    'http_max_multipart_form_data_size',
    'calculate_text_stack_trace',
    'allow_ddl',
    'parallel_view_processing',
    'enable_debug_queries',
    'low_cardinality_allow_in_native_format'
]



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import itertools
import shlex
import logging
import subprocess


from vulkn.clickhouse.client.base import ClickHouseClient
from vulkn.clickhouse import sqlformat
from vulkn.config import FLAGS, SETTINGS, AUTH_OPTS, EXTERNAL_FILE_OPTS
from vulkn.recordset import RecordSet
from vulkn.utils import LogLevels


log = logging.getLogger()


CLI_FLAGS = [
    'disable_suggestion',
    'multiline',
    'multiquery',
    'testmode',
    'vertical',
    'time',
    'stacktrace',
    'progress',
    'echo'
]


CLI_OPTS = {
    'suggestion_limit': 10000
}


def flatten_args(d):
    return list(
        itertools.chain.from_iterable(
            [['--{}'.format(c), str(a)] for c, a in sorted(d.items())]))


def quote_shlex(cmd):
    return ' '.join([shlex.quote(c) for c in cmd])
    

class ClickHouseCLIClient(ClickHouseClient):
    def __init__(self, config_file=None, **kwargs):
        super(ClickHouseCLIClient, self).__init__(config_file, **kwargs)
    
    def _setFlags(self, **kwargs):
        self._flag_opts = filter(
            lambda k: k in set(FLAGS + CLI_FLAGS) and kwargs[k] == True,
            kwargs.keys())

    def _setOptions(self, **kwargs):
        self._opts = dict(
            filter(lambda k: k[0] in CLI_OPTS.keys(), kwargs.items()))

    def _setAuth(self, **kwargs):
        self._auth = dict(
            filter(lambda k: k[0] in AUTH_OPTS.keys(), kwargs.items()))

    def _setExternalFileOptions(self, **kwargs):
        self._external_file_opts = dict(
            filter(lambda k: k[0] in EXTERNAL_FILE_OPTS.keys(), kwargs.items()))

    def _setSettings(self, **kwargs):
        self._settings = dict(
            filter(lambda k: k[0] in SETTINGS, kwargs.items()))

    def _cli(self,
             config_file_overrides=None,
             flag_opts_overrides=None, 
             auth_overrides=None, 
             opts_overrides=None, 
             settings_overrides=None,
             external_file_overrides=None):
        config_file = config_file_overrides or self._config_file
        flag_opts = flag_opts_overrides or self._flag_opts
        opts = {**self._opts, **(opts_overrides or {})}
        auth_opts = {**self._auth, **(auth_overrides or {})}
        settings = {**self._settings, **(settings_overrides or {})}
        external_file_opts = {**self._external_file_opts,
                              **(external_file_overrides or {})}
        cmd = ['/usr/bin/clickhouse-client']
        if config_file:
            cmd += ['--config-file', config_file]
        if len(flag_opts) > 0:
            cmd += ['--{}'.format(c) for c in flag_opts]
        if len(auth_opts) > 0:
            cmd += flatten_args(auth_opts)
        if len(opts) > 0:
            cmd += flatten_args(opts)
        if len(settings) > 0:
            cmd += flatten_args(settings)
        if len(external_file_opts) > 0:
            cmd += flatten_args(external_file_opts)
        return cmd

    def _q(self, cli, query, settings=None):
        if cli is None:
            flag_opts_overrides = {'disable_suggestion': True}
            opts_overrides = {'format': 'TSVWithNamesAndTypes'}
            cli = self._cli(flag_opts_overrides=flag_opts_overrides,
                            opts_overrides=opts_overrides,
                            settings_overrides=settings)
        log.log(5, quote_shlex(cli))
        log.log(LogLevels.SQL, sqlformat(query))
        p = subprocess.run(cli,
                           input=query,
                           stdout=subprocess.PIPE,
                           universal_newlines=True)
        return p

    def execute(self, query, settings=None):
        q = self._q(None, query, settings=settings)
        return q.returncode

    def select(self, query, settings=None):
        flag_opts_overrides = {'disable_suggestion': True}
        opts_overrides = {'format': 'TSVWithNamesAndTypes'}
        cli = self._cli(flag_opts_overrides=flag_opts_overrides,
                        opts_overrides=opts_overrides,
                        settings_overrides=settings)
        q = self._q(cli, query, settings=settings)
        return RecordSet(q.stdout)
        
    def insert(self, query, settings=None):
        c = self._cli(settings=settings)

    def insert_blob(self, data, table, input_format='TSV', settings=None):
        flag_opts_overrides = {'disable_suggestion': True}
        opts_overrides = {'format': 'TSVWithNamesAndTypes'}
        cli = self._cli(flag_opts_overrides=flag_opts_overrides,
                        opts_overrides=opts_overrides,
                        settings_overrides=settings)
        cli += ['--query', 'INSERT INTO {} FORMAT {}'.format(table, input_format)]
        env = {'LC_ALL': 'C'}
        p = subprocess.run(cli, stdout=subprocess.PIPE, input=data, env=env, encoding='ascii')
        return p.returncode



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import re

import sqlparse
import sqlparse.keywords
from sqlparse.sql import TokenList


class Where(sqlparse.sql.TokenList):
    M_OPEN = sqlparse.tokens.Keyword, 'WHERE'
    M_CLOSE = sqlparse.tokens.Keyword, (
        'ORDER BY', 'GROUP BY', 'VECTORIZE BY', 'LIMIT', 'UNION', 'UNION ALL', 'EXCEPT', 'HAVING', 
        'RETURNING', 'INTO')

sqlparse.sql.Where = Where

SQL_REGEX = {
    'clickhouse-ext': [
        (r'VECTORIZE\s+BY\b', sqlparse.tokens.Keyword),
        (r'CHUNK\s+BY\b', sqlparse.tokens.Keyword)
    ]
}

FLAGS = re.IGNORECASE | re.UNICODE
SQL_REGEX = [(re.compile(rx, FLAGS).match, tt) for rx, tt in SQL_REGEX['clickhouse-ext']]

sqlparse.keywords.SQL_REGEX = SQL_REGEX + sqlparse.keywords.SQL_REGEX
sqlparse.lexer.SQL_REGEX = sqlparse.keywords.SQL_REGEX



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.clickhouse.client.cli import ClickHouseCLIClient
from vulkn.clickhouse.client.http import ClickHouseHTTPClient
   

class ClickHouseHybridClient():
    def __init__(self, config_file=None, **kwargs):
        self._cli = ClickHouseCLIClient(config_file, **kwargs)
        self._http = ClickHouseHTTPClient(config_file, **kwargs)
    
    def _q(self, query, settings=None):
        pass

    def execute(self, query, settings=None):
        q = self._q(query, settings=settings)
        return q.returncode

    def select(self, query, settings=None):
        q = self._q(query, settings=settings)
        return RecordSet(q.stdout)
        
    def insert(self, query, settings=None):
        pass

    def insert_blob(self, query, input_format='TSV', settings=None):
        pass

    def create_database(self, database):
        return self.execute(
            'CREATE DATABASE IF NOT EXISTS {database}'.format(database=database))

    def drop_table(self, database, table):
        return self.execute(
            'DROP TABLE IF EXISTS {database}.{table}'.format(database=database, 
                                                             table=table))

    def drop_database(self, database):
        return self.execute(
            'DROP DATABASE IF EXISTS {database}'.format(database=database))

    def drop_view(self, database, view):
        self.drop_table(database, view)

    def fetch_scalar(self, query, column):
        return self.select(query)[0][column]

    def check_bool(self, query, column):
        return int(self.fetch_scalar(query, column)) == 1

    def table_exists(self, database, table):
        return self.check_bool(
            'EXISTS TABLE {database}.{table}'.format(database=database,
                                                     table=table),
            'result')

    def view_exists(self, database, view):
        return self.check_bool("""
            SELECT
                count()>0 AS result
            FROM system.tables
            WHERE 
                database = '{database}' 
                AND name = '{view}' 
                AND engine = 'View'""".format(database=database, view=view),
            'result')

    def database_exists(self, database):
        return self.check_bool("""
            SELECT
                count()>0 AS result
            FROM system.databases
            WHERE name = '{database}'""".format(database=database),
            'result')
    
    def detach_table(self, database, table):
        self.execute('DETACH TABLE {database}.{table}'.format(
            database=database, table=table))

    def list_databases(self):
        return [r['name'] for r in self.select('SHOW DATABASES')]

    def list_tables(self, database):
        return [r['name'] for r in self.select("""
            SELECT
                name
            FROM system.tables
            WHERE create_table_query LIKE 'CREATE TABLE%'
                AND database = '{database}'
            ORDER BY name""".format(database=database))]

    def list_views(self, database):
        return [r['name'] for r in self.select("""
            SELECT
                name
            FROM system.tables
            WHERE engine = 'View' AND database = '{database}'
            ORDER BY name""".format(database=database))]



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import sys
import random
import logging

from vulkn.sql.utils import ast_protect_injection


log = logging.getLogger()


@ast_protect_injection
def vectorWindow(v, backward=None, forward=None):
    varg = random.randint(1,sys.maxsize)
    select_expr = []
    arr_join_expr = []
    if backward is None and forward is None:
        return (f'[v{varg}]', (f'{v} AS v{varg}',))
    backward = int(str(backward or 0))
    forward = int(str(forward or 0))
    if int(str(forward)) < int(str(backward)):
        raise Exception('vectorWindow forward must be greater than backwards.')
    for idx, i in enumerate(range(backward, forward+1)):
        select_expr.append(f"v{varg}_{idx+1}")
        if i < 0:
            arr_join_expr.append(
                f"arrayConcat(arrayWithConstant(abs({i}), NULL), arraySlice(__{v}, 1, {i})) AS v{varg}_{idx+1}")
        elif i == 0:
            arr_join_expr.append(f"__{v} AS v{varg}_{idx+1}")
        else: # i > 0
            arr_join_expr.append(
                f"arrayConcat(arraySlice(__{v}, 1+{i}), arrayWithConstant(abs({i}), NULL)) AS v{varg}_{idx+1}")
    return ("[{}]".format(','.join(select_expr)), arr_join_expr)

@ast_protect_injection
def vectorWindowAgg(agg, v, backward=None, forward=None):
    w = vectorWindow(v, backward, forward)
    varg = w[0]
    return (f"arrayReduce('{agg}', {varg})", w[1])

@ast_protect_injection
def vectorDeltaLag(v, offset=1):
    offset = int(str(offset))
    if offset < 1:
        raise Exception('Parameter "offset" must be greater than 0')
    varg = random.randint(1,sys.maxsize)
    select_expr = f'v{varg}_1 - v{varg}_2'
    arr_join_expr = (f'__{v} AS v{varg}_1',
                     f"arrayConcat(arrayWithConstant({offset}, NULL), arraySlice(__{v}, {offset}, -{offset})) AS v{varg}_2",)
    return (select_expr, arr_join_expr)

@ast_protect_injection
def vectorDeltaLead(v, offset=1):
    offset = int(str(offset))
    if offset < 1:
        raise Exception('Parameter "offset" must be greater than 0')
    varg = random.randint(1,sys.maxsize)
    select_expr = f'v{varg}_2 - v{varg}_1'
    arr_join_expr = (f'__{v} AS v{varg}_1',
                     f"arrayConcat(arraySlice(__{v}, 1+{offset}), arrayWithConstant({offset}, NULL)) AS v{varg}_2",)
    return (select_expr, arr_join_expr)

vectorDelta = vectorDeltaLag

# TODO: Performance problems. Leaving disabled.
# @ast_protect_injection
# def vectorAccumulate(v):
#    varg = random.randint(1,sys.maxsize)
#    return (f'runningAccumulate(v{varg})', (f"arrayReduce('sumStateForEach', [__{v}]) AS v{varg}",))



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import getpass
import os
import yaml
import uuid
import shutil
import subprocess
import time
import pathlib
import atexit


from vulkn.config import VulknConfig
from vulkn.clickhouse.client.cli import ClickHouseCLIClient
from vulkn.utils import get_next_free_socket


templates = '{}/templates'.format(pathlib.Path(__file__).parent.absolute())
CONFIG_TEMPLATE = open(f'{templates}/config.xml').read()
USERS = open(f'{templates}/users.xml').read()


class UserWorkSpaceManager():
    def __init__(self):
        self._username = getpass.getuser()
        self._config_dir = '{}/.vulkn'.format(os.path.expanduser('~'))
        os.makedirs(self._config_dir + '/default', 0o700, exist_ok=True)
        if not os.path.exists(self._config_dir + '/config.yaml'):
            config = {
                'folios': {
                    self._config_dir + '/default': []
                }
            }
            open(self._config_dir + '/config.yaml', 'w').write(yaml.dump(config))

    def get_config(self):
        return yaml.load(open(self._config_dir + '/config.yaml').read(), Loader=yaml.FullLoader)

    def add_workspace(self, uuid: str, workspace: str, folio: str='default', name: str=None):
        config = self.get_config()
        if folio == 'default' or folio is None:
            folio = self._config_dir + '/default'
        if folio in config['folios']:
            for w in config['folios'][folio]:
                if uuid == w['uuid']:
                    return
        config['folios'].setdefault(folio, []).append(
            {'name': name or '', 'uuid': str(uuid), 'workspace': workspace})
        open(self._config_dir + '/config.yaml', 'w').write(yaml.dump(config))

    def remove_workspace(self, uuid: str):
        config = self.get_config()
        for k, v in config['folios'].items():
            config['folios'][k] = [i for i in v if i['uuid'] != uuid]
        open(self._config_dir + '/config.yaml', 'w').write(yaml.dump(config))

    def save_auth(self, server: str, username: str):
        password = getpass.getpass(prompt='Password: ', stream=None) 
        config = self.get_config()
        config.setdefault('auth', {})[server] = {'username': username, 'password': password}
        open(self._config_dir + '/config.yaml', 'w').write(yaml.dump(config))

    def get_workspace(self, workspace, folio, persist):
        uuid_str = None
        if (workspace and os.path.exists(workspace)):
            uuid_str = open(workspace + '/vulkn/uuid').read().strip()
        else:
            uuid_str = str(uuid.uuid4())
        if persist:
            if (folio is None or folio == 'default') and workspace is None:
                workspace = self._config_dir + '/' + uuid_str
        else:
            if workspace is None:
                workspace = '/tmp/{}-{}'.format(getpass.getuser(), uuid_str)
        return (workspace, uuid_str)


class LocalWorkSpace():
    def __init__(self,
                 workspace: str=None,
                 folio: str=None,
                 name: str=None,
                 persist: bool=False,
                 http_port=None,
                 port=None,
                 overwrite=False,
                 timeout=10):
        if port is None:
            port = get_next_free_socket('127.0.0.1', list(range(9001,10000)))
        if http_port is None:
            http_port = get_next_free_socket('127.0.0.1', list(range(8124,8999)))
        self._workspace, self._uuid = UserWorkSpaceManager().get_workspace(workspace, folio, persist)
        self._folio = folio
        self._name = name
        self._persist = persist
        self._timeout = timeout
        self._port = port
        self._http_port = http_port
        self._service = None
        self._config = {
            'log_file': f'{self._workspace}/vulkn/logs/clickhouse-server.log',
            'error_log_file': f'{self._workspace}/vulkn/logs/clickhouse-server.err.log',
            'session_name': 'vulkn',
            'http_port': str(http_port),
            'tcp_port': str(port),
            'data_directory': f'{self._workspace}/vulkn/clickhouse',
            'tmp_directory': f'{self._workspace}/vulkn/clickhouse/tmp',
            'user_files_directory': f'{self._workspace}/vulkn/user_files',
            'format_schema_directory': f'{self._workspace}/vulkn/schemas'
        }

        if os.path.exists(self._workspace):
            if overwrite:
                self._cleanup()
        else:
            os.makedirs(self._workspace, 0o777, exist_ok=False)
            dirs = ['build', 'cores', 'data', 'dictionaries_lib', 'flags', 'format_schemas',
                    'metadata', 'preprocessed_configs', 'storage', 'tmp', 'user_files']
            for d in dirs:
                os.makedirs(f'{self._workspace}/vulkn/clickhouse/{d}', 0o777, exist_ok=True)

        os.makedirs(f'{self._workspace}/vulkn/etc', 0o777, exist_ok=True)

        uuid_info = open(f'{self._workspace}/vulkn/uuid', 'w')
        uuid_info.write(str(self._uuid) + '\n')
        uuid_info.close()

        conf = open(f'{self._workspace}/vulkn/etc/config.xml', 'w')
        conf.write(CONFIG_TEMPLATE.format(**self._config))
        conf.close()

        users = open(f'{self._workspace}/vulkn/etc/users.xml', 'w')
        users.write(USERS)
        users.close()

        self.start()
        atexit.register(self.stop)

    @classmethod
    def tempWorkSpace(cls, path=None):
        w = cls(persist=False, workspace=path)
        return w

    def start(self):
        server = ['clickhouse-server', 
                  '-C', f'{self._workspace}/vulkn/etc/config.xml',
                  '-P', f'{self._workspace}/vulkn/clickhouse.pid']
        self._service = subprocess.Popen(server, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        for i in range(0, self._timeout):
            time.sleep(0.1)
            try:
                log_data = open(f'{self._workspace}/vulkn/logs/clickhouse-server.log', 'r').read()
                if 'Ready for connections' in log_data:
                    cx = ClickHouseCLIClient().setAuth(host='localhost', port=self._config['tcp_port'])
                    cx.execute('CREATE DATABASE IF NOT EXISTS {}'.format(VulknConfig['tmp']))
                    err_log = open(f'{self._workspace}/vulkn/logs/clickhouse-server.err.log', 'r').read()
                    print(err_log)
                    print(log_data)
                    break
            except:
                pass

        if self._persist:
            UserWorkSpaceManager().add_workspace(self._uuid, self._workspace, self._folio, self._name)

    def _cleanup(self):
        shutil.rmtree(f'{self._workspace}/vulkn')
        if not os.listdir(self._workspace):
            os.rmdir(self._workspace)

    def stop(self):
        self._service.terminate()
        try:
            outs, _ = self._service.communicate(timeout=self._timeout)
            print('Exit code', self._service.returncode)
            print(outs.decode('utf-8'))
        except subprocess.TimeoutExpired:
            print('subprocess did not terminate in time')
        if not self._persist:
            self._cleanup()
        else:
            os.rename(
                f'{self._workspace}/vulkn/logs/clickhouse-server.log',
                '{}/vulkn/logs/clickhouse-server.log.{}'.format(self._workspace, str(int(time.time()))))
            os.rename(
                f'{self._workspace}/vulkn/logs/clickhouse-server.err.log',
                '{}/vulkn/logs/clickhouse-server.err.log.{}'.format(self._workspace, str(int(time.time()))))



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.types.base import func
from vulkn.types import *


def count(expression=None, distinct=False):
    if distinct and expression:
        return UInt64(func('countDistinct', expression))
    if expression:
        return UInt64(func('count', expression))
    return UInt64(func('count'))

def any(x):
    return col(func('any', x))

def anyHeavy(x):
    return col(func('anyHeavy', x))

def anyLast(x):
    return col(func('anyLast', x))

def groupBitAnd(expr):
    return UInt64(func('groupBitAnd', expr))

def groupBitOr(expr):
    return UInt64(func('groupBitOr', expr))

def groupBitXor(expr):
    return UInt64(func('groupBitXor', expr))

def groupBitmap(expr):
    return UInt64(func('groupBitmap', expr))

def min(x):
    return col(func('min', x))

def max(x):
    return col(func('max', x))

def argMin(arg, val):
    return col(func('argMin', arg, val))

def argMax(arg, val):
    return col(func('argMax', arg, val))

def sum(x):
    return col(func('sum', x))

def sumWithOverflow(x):
    return col(func('sumWithOverflow', x))

def sumMap(key, value):
    return col(func('sumMap', key, value))

def skewPop(x):
    return Float64(func('skewPop', x))

def skewSamp(x):
    return Float64(func('skewSamp', x))

def kurtPop(x):
    return Float64(func('any', x))

def kurtSamp(x):
    return Float64(func('kurtSamp', x))

def timeSeriesGroupSum(uid, timestamp, value):
    return Array(func('timeSeriesGroupSum', uid, timestamp, value))

def timeSeriesGroupRateSum(uid, ts, val):
    return Array(func('timeSeriesGroupRateSum', uid, ts, val))

def avg(x):
    return Float64(func('avg', x))

def uniq(*args):
    return UInt64(func('uniq', *args))

def uniqCombined(*args):
    return UInt64(func('uniqCombined', *args))

def uniqHLL12(*args):
    return UInt64(func('uniqHLL12', *args))

def uniqExact(*args):
    return UInt64(func('uniqExact', *args))

def groupArray(x, max_size=None):
    if max_size:
        return Array(func('groupArray({})'.format(max_size), x))
    return Array(func('groupArray', x))

def groupArrayInsertAt(position, x):
    return Array(func('groupArrayInsertAt', position, x))

def groupArrayMovingSum(x, window_size=None):
    if window_size:
        Array(func('groupArrayMovingSum({})'.format(window_size), x))
    return Array(func('groupArrayMovingSum', x))

def groupArrayMovingAvg(x, window_size=None):
    if window_size:
        Array(func('groupArrayMovingAvg({})'.format(window_size), x))
    return Array(func('groupArrayMovingAvg', x))

def groupUniqArray(x, max_size=None):
    if max_size:
        return Array(func('groupUniqArray({})'.format(max_size), x))
    return Array(func('groupUniqArray', x))

def quantile(level, x):
    return col(func('quantile({})'.format(level), x))

def quantileDeterministic(level, x, determinator):
    return col(func('quantileDeterministic', level, x, determinator))

def quantileTiming(level, expression):
    return col(func('quantile({})'.format(level), expression))

def quantileTimingWeighted(level, x, weight):
    return col(func('quantileTimingWeighted', level, x, weight))

def quantileExact(level, x):
    return col(func('quantileExact({})'.format(level), x))

def quantileExactWeighted(level, x, weight):
    return col(func('quantileExactWeighted', level, x, weight))

def quantileTDigest(level, x):
    return col(func('quantileTDigest({})'.format(level), x))

def median(x):
    return col(func('median', x))

# TODO Next release. Add median variants.

def quantiles(levels, x):
    return Array(func('quantiles', *levels, x))

# TODO Next release. Add quantiles variants.

def varSamp(x):
    return Float64(func('varSamp', x))

def varPop(x):
    return Float64(func('varPop', x))

def stddevSamp(x):
    return Float64(func('stddevSamp', x))

def stddevPop(x):
    return Float64(func('stddevPop', x))

def topK(N, column):
    return col(func('topK({})'.format(N), column))

def covarSamp(x, y):
    return Float64(func('covarSamp', x, y))

def covarPop(x, y):
    return Float64(func('covarPop', x, y))

def corr(x, y):
    return Float64(func('corr', x, y))

def histogram(bins, values):
    return Array(func('histogram', bins, values))

def sequenceMatch(pattern, time, conditions):
    return col(func('sequenceMatch', pattern, time, *conditions))

def sequenceCount(pattern, time, conditions):
    return UInt64(func('sequenceCount', pattern, time, *conditions))

def windowFunnel(window, timestamp, conditions):
    return UInt64(func('windowFunnel', window, timestamp, *conditions))

def retention(*conditions):
    return UInt8(func('retention', *conditions))

def uniqUpTo(N, x):
    return UInt64(func('uniqUpTo', N, x))

def sumMapFiltered(keys_to_keep, keys, values):
    return col(func('sumMapFiltered', keys_to_keep, keys, values))

"""
class AggregateFunction:
    def __init__(self, *args):
        pass

    @classmethod
    def If(self, cond):
        r = AggregateFunction(self._func, self._args)
        r._if = True
        return r

    @classmethod
    def Array(self):
        r = AggregateFunction(self._func, self._args)
        r._array = True
        return r

    @classmethod
    def State(self):
        r = AggregateFunction(self._func, self._args)
        r._state = True
        return r

    @classmethod
    def Merge(self):
        r = AggregateFunction(self._func, self._args)
        r._merge = True
        return r
    
    @classmethod    
    def MergeState(self):
        r = AggregateFunction(self._func, self._args)
        r._merge_state = True
        return r

    @classmethod
    def ForEach(self):
        r = AggregateFunction(self._func, self._args)
        r._for_each = True
        return r

    @classmethod
    def Resample(self, start, end, stop):
        r = AggregateFunction(self._func, self._args)
        r._resample = True
        return r

    def __str__(self):
        return str(self._value)
"""


# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.types.base import func
from vulkn.types.array import Array
from vulkn.types.string import String
from vulkn.types.integer import UInt64


class URL(String):
    _METHODS = {
        'protocol': None,
        'domain': None,
        'domainWithoutWWW': None,
        'topLevelDomain': None,
        'firstSignificantSubdomain': None,
        'cutToFirstSignificantSubdomain': None,
        'path': None,
        'pathFull': None,
        'queryString': None,
        'fragment': None,
        'queryStringAndFragment': None,
        'extractURLParameters': Array,
        'extractURLParameterNames': Array,
        'URLHierarchy': Array,
        'URLPathHierarchy': Array,
        'decodeURLComponent': Array,
        'cutWWW': None,
        'cutQueryString': None,
        'cutFragment': None,
        'cutQueryStringAndFragment': None,
    }

    def extractURLParameter(self, name):
        return URL(func('extractURLParameter', self._value, name))

    def cutURLParameter(self, name):
        return URL(func('cutURLParameter', self._value, name))

    def URLHash(self, N=None):
        if N:
            return UInt64(func('cutURLParameter', self._value, N))
        else:
            return UInt64(func('cutURLParameter', self._value))


# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import uuid
import weakref
import hashlib
import logging


import vulkn.engines
from vulkn.utils import timer, LogLevels


timing = False
enable_parser = True


log = logging.getLogger()
logging.addLevelName(LogLevels.SQL, 'SQL')


class CacheMode:
    Table=1
    Normal=2
    Array=3
    Merge=4


class TTL:
    Session=1
    

class ClickHouseEnvironment:
    def __init__(self, ctx):
        self._ctx = ctx

    @property
    def settings(self):
        sql = 'SELECT name, value, changed FROM system.settings ORDER BY changed = 1 DESC, name'
        return self._ctx.q(sql).exec()


class VulknSession:
    def __init__(self, ctx, cache_location='vulkn'):
        self._ctx = ctx
        self._cache_location = cache_location
        self._uuid = uuid.uuid4().hex
        self._cache = []
        self._needs_optimize = True
        self.createSession()
        self._finalizer = weakref.finalize(self, self._destroyCache, ctx, self._cache)

    def optimize(self, final=False):
        if self._needs_optimize:
            self._ctx._conn.execute(
                'OPTIMIZE TABLE {session_store} {final}'.format(session_store=self.session_store,
                                                                final='FINAL' if final else ''))
            if final:
                self._needs_optimize = False

    @property
    def session_store(self):
        return '{cache_location}.session_{uuid}'.format(cache_location=self._cache_location,
                                                        uuid=self._uuid)

    def createSession(self):
        cache_ref = self.session_store
        cache_location = self._cache_location
        ctx = self._ctx
        cache_source_sql = f"""
            CREATE TABLE {cache_ref} (
                t_ts DateTime CODEC(LZ4)
            ) ENGINE = MergeTree()
            ORDER BY tuple()"""
        ctx.exec(f'CREATE DATABASE IF NOT EXISTS {cache_location}')
        ctx.exec(f'DROP TABLE IF EXISTS {cache_ref}')
        ctx.exec(cache_source_sql)
        ctx.exec(f'INSERT INTO {cache_ref} VALUES (now())')
        self._cache.append(VulknCachedObject(cache_ref, cache_source_sql))

    def destroySession(self):
        self._finalizer()

    @staticmethod
    def _destroyCache(ctx, cache):
        log.info('Purging cache...')
        for c in cache:
            if isinstance(c, VulknCachedObject):
                if c.ttl == TTL.Session and c.mode == CacheMode.Table:
                    ctx._conn.execute('DROP TABLE IF EXISTS {}'.format(c.ref))
            else:
                ctx._conn.execute('DROP TABLE IF EXISTS {}'.format(c))

    def _create_table_cache(self, obj_hash, obj_query, engine):
        ctx = self._ctx
        cache_location = self._cache_location
        cache_tmp = 'session_{uuid}_v_{obj_hash}'.format(uuid=self._uuid, obj_hash=obj_hash)
        ctx._conn.execute(f'DROP TABLE IF EXISTS {cache_location}.{cache_tmp}')
        ctx._conn.execute(f'CREATE TABLE {cache_location}.{cache_tmp} ENGINE = {engine} AS {obj_query}')
        return VulknCachedObject(f'{cache_location}.{cache_tmp}', obj_query)

    def _store_array_vector(self, obj_hash, obj_query):
        ctx = self._ctx
        cache_location = self._cache_location
        cache_tmp = 'session_{uuid}_v_{obj_hash}'.format(uuid=self._uuid, obj_hash=obj_hash)
        ctx._conn.execute(f'DROP TABLE IF EXISTS {cache_location}.{cache_tmp}')
        ctx._conn.execute(f"""
            CREATE TABLE {cache_location}.{cache_tmp}
                ENGINE = Join(ANY, LEFT, r)
            AS 
            SELECT
                1 AS r
                , v
            FROM ({obj_query})""")
        return VulknCachedObject(f'{cache_location}.{cache_tmp}', obj_query)

    def _store_object(self, obj_hash, obj_query, optimize):
        session_store = self.session_store
        session_column_name = 'var_{}'.format(obj_hash)
        self._ctx._conn.execute(f"""
            ALTER TABLE {session_store} 
                ADD COLUMN IF NOT EXISTS {session_column_name}
                    MATERIALIZED {obj_query} CODEC(LZ4)""")
        self._needs_optimize = True
        if optimize:
            self.optimize(final=True)
        return VulknCachedObject(session_column_name, obj_query, mode=CacheMode.Normal)

    def cache(self, obj, mode=CacheMode.Table, engine=None, optimize=False):
        obj_hash = ''
        obj_query = ''
        if engine is None:
            engine = vulkn.engines.Memory()
        if isinstance(obj, str):
            obj_hash = hashlib.md5(obj.encode()).hexdigest()
            obj_query = obj
        if isinstance(obj, VulknCachedObject):
            return obj
        if mode == CacheMode.Table:
            intermediate_cache = self._create_table_cache(obj_hash, obj_query, engine)
            self._cache.append(intermediate_cache)
            return intermediate_cache.ref
        if mode == CacheMode.Array:
            intermediate_cache = self._store_array_vector(obj_hash, obj_query)
            self._cache.append(intermediate_cache)
            return intermediate_cache.ref
        stored_object = self._store_object(obj_hash, obj_query, optimize)
        return stored_object.ref


class VulknCachedObject:
    def __init__(self, ref, source, ttl=TTL.Session, mode=CacheMode.Table):
        self.ref = ref
        self.source = source
        self.ttl = ttl
        self.mode = mode

    def column_info(self, ctx):
        table_name = self.ref
        sql = f"SELECT name, type FROM system.columns WHERE database || '.' || table = '{table_name}'"
        return ctx.select(sql).to_record()



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import copy
import csv
import subprocess
import logging


import vulkn
from vulkn.utils import LogLevels


log = logging.getLogger()


class CSV:
    CLICKHOUSE_FORMAT = 'CSV'

    OPTIONS = {
        'delimiter': ',',
        'allow_single_quotes': True,
        'allow_double_quotes': True,
        'unquoted_null_literal_is_null': True,
        'input_format_defaults_for_omitted_fields': False,
        'header': False,
        'infer_schema': False,
        'allow_enums': False,
        'column_format': 'snake_case',
        'overwrite': False,
        'engine': None,
        'sample_engine': None,
        'sample_size': 50000,
        'sample_preload': True,
        'schema': None}

    def __init__(self):
        self._options = copy.deepcopy(self.OPTIONS)

    def options(self, **kwargs):
        self._options.update(kwargs)
        return self

    def columns(self, uri):
        cols = []
        with open(uri.path, 'r') as f:
            reader = csv.reader(f)
            cols = next(reader)
        if not self._options['header']:
            cols = ['col{}'.format(i+1) for i, v in enumerate(cols)]
        return cols

    def sample(self, uri):
        lines = []
        i = 0
        with open(uri.path, 'r') as f:
            while True:
                if i == 0 and self._options['header']:
                    i = i+1
                    f.readline()
                    continue
                l = f.readline()
                if l == '' or i >= self._options['sample_size']:
                    break
                lines.append(l)
                i = i+1
        return ''.join(lines)

    def read(self, uri, database, table):
        # TODO: Next release. Hacky. Remove subprocess/cat pipeline.
        v = vulkn.Vulkn()
        header = 'CSVWithNames' if self._options['header'] else 'CSV'
        env = {'LC_ALL': 'C'}
        src = subprocess.Popen(['cat', uri.path], stdout=subprocess.PIPE, env=env, encoding='ascii')
        tgt = ['clickhouse-client', '-A', '-m', '-n']
        tgt += ['--host', v._host, '--port', str(v._port), '--user', v._user, '--password', v._password]
        tgt += ['--query', f'INSERT INTO {database}.{table} FORMAT {header}']
        log.debug(str(tgt))
        log.log(LogLevels.SQL, f'INSERT INTO {database}.{table} FORMAT {header}')
        p = subprocess.Popen(tgt, stdin=src.stdout, stdout=subprocess.PIPE, env=env, encoding='ascii')
        src.stdout.close()
        p.communicate()

    def write(self):
        pass




# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import ast
import logging


import vulkn.sql
import vulkn.table
import vulkn.reader
import vulkn.datatable
from vulkn.utils import singleton
from vulkn.catalog import VulknCatalogViewer
from vulkn.data import VulknDataStore
from vulkn.session import VulknSession
from vulkn.database import VulknClickHouseDatabaseMixIn, MutationManager, VulknSystemManager
from vulkn.scheduler import VulknScheduler
from vulkn.clickhouse.client.cli import ClickHouseCLIClient


log = logging.getLogger()


class VulknDataTablesMixIn:
    def q(self, query):
        return vulkn.datatable.QueryStringDataTable(self, query)

    def select(self, *cols):
        return vulkn.datatable.SelectQueryDataTable(self).select(*cols)

    def s(self, *cols):
        r = self.select(*cols).exec().to_list()
        if len(r) > 0:
            try:
                return ast.literal_eval(r[0] if len(r[0]) > 1 else r[0][0])
            except:
                return r[0] if len(r[0]) > 1 else r[0][0]
        return None

    def numbers(self, count, system=False, mt=False):
        return vulkn.datatable.NumbersDataTable(self, count, system, mt)

    def range(self, start, end, system=False, mt=False):
        return vulkn.datatable.RangeDataTable(self, start, end, system, mt)

    def one(self):
        return vulkn.datatable.OneDataTable(self)

    def random(self, count, start=0, end=18446744073709551615, system=False, mt=False):
        return vulkn.datatable.RandomDataTable(self, count, start, end, system, mt)

    def randfloat(self, count, start=0, end=18446744073709551615, system=False, mt=False):
        return vulkn.datatable.RandomFloatDataTable(self, count, start, end, system, mt)
    
    def with_(self, *cols):
        return vulkn.datatable.SelectQueryDataTable(self).with_(*cols)

    def update(self, table):
        return vulkn.datatable.UpdateQueryDataTable(self, table)

    def delete(self, table):
        return vulkn.datatable.DeleteQueryDataTable(self, table)


@singleton
class Vulkn(VulknDataTablesMixIn, VulknClickHouseDatabaseMixIn):
    def __init__(self,
                 user='default',
                 password='',
                 host='localhost',
                 port=9000,
                 database='default',
                 workspace=None):
        if workspace is not None:
            port = workspace._port
            http_port = workspace._http_port
        self._user = user
        self._password = password
        self._host = host
        self._port = port
        self._database = database
        self._conn = self._reload_conn()
        self.session = None
        self.scheduler = VulknScheduler(self)
        self.session = VulknSession(self)
        self.mutations = MutationManager(self)
        self.read = vulkn.reader.Reader()
        self.table = vulkn.table.Table
        self.system = VulknCatalogViewer(self,
                                         catalog='system',
                                         hidden=['numbers','numbers_mt','one','zero','zero_mt'])
        self.data = VulknDataStore(self)
        self.sys = VulknSystemManager(self)

    def _reload(self):
        if self.session:
            self.session.destroySession()
            del self.session
        self._conn = self._reload_conn()
        self.session = VulknSession(self)

    def _reload_conn(self):
        return ClickHouseCLIClient().setAuth(host=self._host,
                                             port=self._port,
                                             user=self._user,
                                             password=self._password,
                                             database=self._database)




# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from functools import partial


from vulkn.types.base import quote_literal, func, TypeBase, Literal
from vulkn.types import String, col, c
import vulkn.funcs.bitmap as bitmap
import vulkn.funcs.encode as encode
import vulkn.funcs.geo as geo
import vulkn.funcs.agg as agg
import vulkn.funcs.vector as vector


def coalesce(*args):
    return c('coalesce({})'.format(', '.join(map(str, args))))

def and_(*args):
    return c('({})'.format(' AND '.join(map(str, args))))

def or_(*args):
    return c('({})'.format(' OR '.join(map(str, args))))

def not_(args):
    return c('NOT ({})'.format(str(args)))

def xor_(*args):
    return c('({})'.format(' XOR '.join(map(str, args))))

def if_(condition, then_arg, else_arg):
    return c('if({}, {}, {})'.format(condition, quote_literal(then_arg), quote_literal(else_arg)))

def multiIf(branches, default):
    r = 'multiIf({branches}, {default})'
    b_stmts = []
    for branch in branches:
        b_stmts.append('{}, {}'.format(branch[0], quote_literal(branch[1])))
    return c(r.format(branches=', '.join(b_stmts), default=quote_literal(default)))

switch = multiIf

def arrayJoin(arr):
    return c('arrayJoin({})'.format(str(arr)))

def arrayGeneric(func_name, *args):
    lambda_exp = (args[0] if (callable(args[0]) or hasattr(args[0], '__call__')) else None)
    params = (None if lambda_exp is None else '({})'.format(', '.join(lambda_exp.__code__.co_varnames)))
    arrs = args if lambda_exp is None else args[1:]
    arrs = ', '.join([str(a) for a in arrs])
    r = ('{func_name}({arrs})' if lambda_exp is None else '{func_name}({params} -> {lambda_exp}, {arrs})')
    s = ''
    if lambda_exp:
        s = r.format(func_name=func_name, params=params, lambda_exp=lambda_exp(), arrs=arrs)
    else:
        s = r.format(func_name=func_name, arrs=arrs)
    return c(s)

arrayMap = partial(arrayGeneric, 'arrayMap')
arrayFilter = partial(arrayGeneric, 'arrayFilter')
arrayCount = partial(arrayGeneric, 'arrayCount')
arrayExists = partial(arrayGeneric, 'arrayExists')
arrayAll = partial(arrayGeneric, 'arrayAll')
arraySum = partial(arrayGeneric, 'arraySum')
arrayFirst = partial(arrayGeneric, 'arrayFirst')
arrayFirstIndex = partial(arrayGeneric, 'arrayFirstIndex')
arrayCumSum = partial(arrayGeneric, 'arrayCumSum')
arrayCumSumNonNegative = partial(arrayGeneric, 'arrayCumSumNonNegative')
arraySort = partial(arrayGeneric, 'arraySort')
arrayReverseSort = partial(arrayGeneric, 'arrayReverseSort')

def formatReadableSize(bytes):
    return String(func('formatReadableSize', bytes))



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.types.base import func, TypeBase, Literal
from vulkn.types import UInt16, UInt8, UInt32


class DateTimeBase(TypeBase):
    _METHODS = {
        'toTimeZone': None
    }

    @staticmethod
    def _with_timezone(*args, timezone):
        return (*args, Literal("'{}'".format(timezone))) if timezone else args

    def toStartOfInterval(self, interval, timezone=None):
        return type(self)(func('toStartOfInterval',
                               *DateTimeBase._with_timezone(self,
                                                            Literal('INTERVAL ' + interval),
                                                            timezone=timezone)))

    time_bucket = toStartOfInterval

    def toStartOfWeek(self, mode=0):
        return Date(func('toStartOfWeek', self, mode))

    def toYearWeek(self, mode=0, timezone=None):
        return type(self)(func('toYearWeek', *DateTimeBase._with_timezone(self, mode, timezone=timezone)))

    def toWeek(self, mode=0, timezone=None):
        return type(self)(func('toWeek', *DateTimeBase._with_timezone(self, mode, timezone=timezone)))

    def toYear(self, timezone=None):
        return UInt16(func('toYear', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toQuarter(self, timezone=None):
        return UInt8(func('toQuarter', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toMonth(self, timezone=None):
        return UInt8(func('toMonth', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toDayOfYear(self, timezone=None):
        return UInt16(func('toDayOfYear', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toDayOfMonth(self, timezone=None):
        return UInt8(func('toDayOfMonth', *DateTimeBase._with_timezone(self, timezone=timezone)))
        
    def toDayOfWeek(self, timezone=None):
        return UInt8(func('toDayOfWeek', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toStartOfMonth(self, timezone=None):
        return type(self)(func('toStartOfMonth', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toStartOfQuarter(self, timezone=None):
        return type(self)(func('toStartOfQuarter', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toStartOfYear(self, timezone=None):
        return type(self)(func('toStartOfYear', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toStartOfISOYear(self, timezone=None):
        return type(self)(func('toStartOfISOYear', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def formatDateTime(self, formatstr, timezone):
        return String(func('toStartOfISOYear', *DateTimeBase._with_timezone(self, formatstr, timezone=timezone)))

    def dateDiff(self, unit, other, timezone=None):
        return type(self)(func('dateDiff', *DateTimeBase._with_timezone(unit, self, other, timezone=timezone)))

    def toYYYYMM(self, timezone=None):
        return type(self)(func('toYYYYMM', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toYYYYMMDD(self, timezone=None):
        return type(self)(func('toYYYYMMDD', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toYYYYMMDDhhmmss(self, timezone=None):
        return type(self)(func('toYYYYMMDDhhmmss', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toISOWeek(self, timezone=None):
        return type(self)(func('toISOWeek', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toISOYear(self, timezone=None):
        return type(self)(func('toISOYear', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toMonday(self, timezone=None):
        return Date(func('toMonday', self))


class Date(DateTimeBase):
    TYPE = 'Date'
    CAST = 'Date'

    def __str__(self) -> str:
        if type(self._value) is Literal:
            return str(self._value)
        if type(self._value) is func:
            return str(self._value)
        return str("toDate('{}')".format(self._value))

    @classmethod
    def yesterday(self):
        return Date(Literal('yesterday()'))

    @classmethod
    def today(self):
        return Date(Literal('today()'))


class DateTime(DateTimeBase):
    TYPE = 'DateTime'
    CAST = 'DateTime'

    def __str__(self) -> str:
        if type(self._value) is Literal:
            return str(self._value)
        if type(self._value) is func:
            return str(self._value)
        return str("toDateTime('{}')".format(self._value))

    @classmethod
    def now(self):
        return DateTime(Literal('now()'))

    @staticmethod
    def timeSlots(start_time, duration, size):
        return DateTime(func('timeSlots', self, duration, size))

    def toStartOfYear(self, timezone=None):
        return Date(func('toStartOfYear', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toStartOfQuarter(self, timezone=None):
        return Date(func('toStartOfQuarter', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toStartOfMonth(self, timezone=None):
        return Date(func('toStartOfMonth', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toStartOfWeek(self, mode=0, timezone=None):
        return Date(func('toStartOfWeek', *DateTimeBase._with_timezone(self, mode, timezone=timezone)))

    def toStartOfDay(self, timezone=None):
        return DateTime(func('toStartOfDay', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toStartOfHour(self, timezone=None):
        return DateTime(func('toStartOfHour', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toStartOfFifteenMinutes(self, timezone=None):
        return DateTime(func('toStartOfFifteenMinutes', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toStartOfTenMinutes(self, timezone=None):
        return DateTime(func('toStartOfTenMinutes', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toStartOfFiveMinute(self, timezone=None):
        return DateTime(func('toStartOfFiveMinute', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toStartOfMinute(self, timezone=None):
        return DateTime(func('toStartOfMinute', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toUnixTimestamp(self, timezone=None):
        return UInt32(func('toUnixTimestamp', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def timeSlot(self, timezone=None):
        return DateTime(func('timeSlot', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toTime(self, timezone=None):
        return DateTime(func('toTime', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toHour(self, timezone=None):
        return UInt8(func('toHour', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toMinute(self, timezone=None):
        return UInt8(func('toMinute', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toSecond(self, timezone=None):
        return UInt8(func('toSecond', *DateTimeBase._with_timezone(self, timezone=timezone)))

    def toMonday(self, timezone=None):
        return Date(func('toMonday', *DateTimeBase._with_timezone(self, timezone=timezone)))


# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import logging
from typing import List


from vulkn.contrib.sqlparse import sqlparse
from vulkn.sql.statement import SQLStatement


log = logging.getLogger()


class SQLMessage:
    def __init__(self, message: str) -> None:
        self._original_message = message

    def statements(self) -> List:
        return [SQLStatement(s) for s in sqlparse.split(self._original_message)]

    @property
    def original_message(self) -> str:
        return self._original_message

    def __str__(self) -> str:
        return '\n'.join([str(s) for s in self.statements()])



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.types.base import func
from vulkn.types.array import Array
from vulkn.types.string import String
from vulkn.types.integer import UInt32, UInt64


class IP(String):
    def IPv4CIDRToRange(self, cidr):
        return tuple(func('IPv4CIDRToRange', cidr))

    def IPv6CIDRToRange(self, cidr):
        return tuple(func('IPv6CIDRToRange', cidr))

    @staticmethod
    def IPv4NumToString(num):
        return String(func('IPv4NumToString', num))

    @staticmethod
    def IPv6NumToString(num):
        return String(func('IPv6NumToString', num))

    @staticmethod
    def IPv4NumToStringClassC(num):
        return tuple(func('IPv4NumToStringClassC', num))

    @staticmethod
    def IPv6StringToNum(s):
        return String(func('IPv6StringToNum', s))

    @staticmethod
    def IPv4StringToNum(s):
        return UInt32(func('IPv4StringToNum', s))

    @staticmethod
    def IPv4ToIPv6(x):
        return String(func('IPv4ToIPv6', x))

    def cutIPv6(self, bitsToCutForIPv6, bitsToCutForIPv4):
        return String(func('cutIPv6', bitsToCutForIPv6, bitsToCutForIPv4))

    toIPv4 = IPv4StringToNum
    toIPv6 = IPv6StringToNum

    @staticmethod
    def MACNumToString(num):
        return String(func('MACNumToString', num))

    @staticmethod
    def MACStringToNum(mac):
        return UInt64(func('MACStringToNum', mac))

    @staticmethod
    def MACStringToOUI(mac):
        return UInt64(func('MACStringToOUI', mac))


# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.types.base import Literal
from vulkn.types.string import String


class UUID(String):
    TYPE = 'String'
    CAST = 'String'
    _METHODS = {
        'UUIDStringToNum': 'String',
        'UUIDNumToString': 'String'
    }

    @staticmethod
    def generateUUIDv4():
        return UUID(Literal('generateUUIDv4()'))

    UUIDv4 = generateUUIDv4


# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.types.base import TypeBase, Literal, func, FunctionExpression, quote_literal
from vulkn.utils import item_index


class Array(TypeBase):
    _METHODS = {
        'length': 'String'
    }

    _HIGHER_ORDER = [
        'map', 'filter', 'count', 'exists', 'all', 'sum', 
        'first', 'first_index', 'cum_sum', 'sort','reverse_sort'
    ]

    def __getitem__(self, index):
        from vulkn.types.string import String

        start, stop = item_index(index)
        if stop is None:
            return String(func('arrayElement', self._value, start))
        else:
            return String(func('arraySlice', self._value, start, stop))

    def __getattr__(self, name: str):
        def _method(lambda_arg):
            func = 'array{}'.format(''.join(map(str.title, name.split('_'))))
            v = FunctionExpression(func, Literal(lambda_arg), quote_literal(self._value))
            return Array(v)

        if name in [m for m in self._HIGHER_ORDER]:
            return _method
        else:
            return super(Array, self).__getattr__(name)

    def arrayStringConcat(self, separator: str=''):
        from vulkn.types.string import String
        return String(func('arrayStringConcat', self._value, separator))

    join = arrayStringConcat



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.types.base import func
from vulkn.types import String, Array


def hex(arg):
    return String(func('hex', arg))

def unhex(arg):
    return String(func('unhex', arg))

def UUIDStringToNum(arg):
    return String(func('UUIDStringToNum', arg))

def UUIDNumToString(arg):
    return String(func('UUIDNumToString', arg))

def bitmaskToList(arg):
    return String(func('bitmaskToList', arg))

def bitmaskToArray(arg):
    return Array(func('bitmaskToArray', arg))



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import pprint
import logging
import itertools
from typing import List
from functools import wraps


from vulkn.contrib.sqlparse import sqlparse


log = logging.getLogger()


def validate_parameters(*args):
    safe_tokens = set([sqlparse.tokens.Name])
    for a in args:
        if hasattr(a, 'tokens'):
            if not set([t.ttype for t in a.tokens]).issubset(safe_tokens):
                return False
    return True


# TODO: This function protects against SQL injection as far as Vulkn UDFs and other internal query
# rewriting is concerned however as the Vulkn SQL engine is a second level engine it doesn't handle
# cases where users manually construct poisonous strings which are used to build a query. In this
# case the SQL injection is present *before* it is submitted to the Vulkn rewriter. A future Vulkn 
# SQL Server will handle SQL injection (as it will be responsible for building the query at that
# point however is out of scope for the initial release).
def ast_protect_injection(function):
    @wraps(function)
    def protected_call(*args):
        if not validate_parameters(*args):
            raise Exception('Cancelling due to potential SQL injection')
        return function(*args)
    return protected_call


def ast_where_clause(ast):
    (idx, where_clause) = ast.token_next_by(i=sqlparse.sql.Where)
    return where_clause


def ast_from_clause(ast):
    for idx, token in enumerate(ast):
        if token.ttype == sqlparse.tokens.Keyword and token.normalized == 'FROM':
            return ast.token_next(idx)[1]
    else:
        return None


def ast_column_projections(ast):
    (start_idx, ident_list) = ast.token_next_by(i=sqlparse.sql.IdentifierList)
    return list(ident_list.get_identifiers())


def get_keyword_function_args(q, start_idx: int) -> List:
    arg_clause = [l for l in q[start_idx+1:] if not l.is_whitespace][0]
    return list(map(str, list(arg_clause[1].get_identifiers())))


def del_keyword_function_node(tokens, q, start_idx: int) -> str:
    for i in range(start_idx, len(q.tokens)):
        if q[i].is_group:
            del tokens[i]
            break
        else:
            del tokens[i]
    return str(q).strip()


def validate_statement_excludes(excludes, statement) -> tuple:
    funcs = lambda tk: sqlparse.utils.imt(tk, m=[(sqlparse.tokens.Keyword, e) for e in excludes])
    return statement._token_matching(funcs) == (None, None)
   

def get_functions(tokens):
    if isinstance(tokens, sqlparse.sql.Function):
        yield tokens[0].value
    for idx, token in enumerate(tokens):
        if isinstance(token, sqlparse.sql.Function):
            yield token[0].value
        if token.is_group:
            for g in get_functions(token):
                yield g


def get_identifiers(tokens):
    for idx, token in enumerate(tokens):
        if isinstance(token, sqlparse.sql.Identifier) and type(token.parent) != sqlparse.sql.Function:
            yield token[0].value
        if token.is_group:
            for g in get_identifiers(token):
                yield g


def resolve_identifiers(identifiers):
    ids = []
    seq = itertools.count(1)
    for l in identifiers:
        if type(l) == sqlparse.sql.Token:
            ids.append({'node': l, 'src': [], 'funcs': [], 'id': l.value})
        else:
            if len(l.tokens) == 1:
                ids.append({'node': l, 'src': [l.value], 'funcs': [], 'id': l.get_real_name()})
            else:
                node = {'node': l, 'id': l.get_alias() or str(l)}
                node['src'] = list(set(filter(lambda x: x != node['id'], get_identifiers(l))))
                node['funcs'] = list(set(list(get_functions(l))))
                ids.append(node)
    log.debug(ids)
    return ids



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


class ClickHouseCLIWriter:
    OPTIONS = { 'options': {}, 'client': None, 'format': None}

    def __init__(self):
        self._options = copy.deepcopy(self.OPTIONS)

    def options(self, **kwargs):
        self._options.update(kwargs)
        return self

    def _cli(self, database, table):
        flag_opts_overrides = {'disable_suggestion': True}
        opts_overrides = {'format': self._options['format'] or 'CSV'}
        cli = self._cli(flag_opts_overrides=flag_opts_overrides,
                        opts_overrides=opts_overrides,
                        settings_overrides=settings)
        cli += ['--query', 'INSERT INTO {}.{} FORMAT {}'.format(database, table, input_format)]
        env = {'LC_ALL': 'C'}
        return cli
        
    def write(self, source, database, table):
        target = self._cli(database, table)
        p = subprocess.run(
            cli, stdout=subprocess.PIPE, input=source.buffer(), env=env, encoding='ascii')
        return p.returncode


# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.types.base import func
from vulkn.types import UInt64, UInt8, Array, col


def vectorWindow(v, backward=None, forward=None):
    return col(func('vectorWindow', v, backward, forward))

def vectorWindowAgg(agg, v, backward=None, forward=None):
    return col(func('vectorWindowAgg', v, backward, forward))

def vectorDeltaLag(v, offset=1):
    return col(func('vectorDeltaLag', v, offset))

def vectorDeltaLead(v, offset=1):
    return col(func('vectorDeltaLead', v, offset))

vectorDelta = vectorDeltaLag

def vectorRowNumber():
    return col(func('vectorRowNumber'))

vectorRowNum = vectorRowNumber

def vectorAgg(agg, v):
    return col(func('vectorAgg', agg, v))

def vectorFirst(v):
    return col(func('vectorFirst', v))

def vectorLast(v):
    return col(func('vectorLast', v))

def vectorLag(v, offset=-1):
    return col(func('vectorLag', v, offset))

def vectorLead(v, offset=1):
    return col(func('vectorLead', v, offset))

def vectorDenseRank(v):
    return col(func('vectorDenseRank', v))

def vectorCumeDist(v):
    return col(func('vectorCumeDist', v))

def vectorPercentRank(v):
    return col(func('vectorPercentRank', v))

def vectorMap(func, v):
    return col(func('vectorMap', func, v))

def vectorNth(v, offset=1):
    return col(func('vectorNth', v, offset))

def vectorNtile(v, buckets):
    return col(func('vectorNtile', v, buckets))

def vectorRank(v):
    return col(func('vectorRank', v))



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.types.base import TypeBase, func, Literal
from vulkn.types.array import Array
from vulkn.types.integer import UInt8, UInt64, Int64
from vulkn.types.float import Float64
from vulkn.types.string import String


class JSON(TypeBase):
    TYPE = 'String'
    CAST = 'String'

    def visitParamHas(self, name):
        return UInt8(func('visitParamHas', self._value, name))

    def visitParamExtractUInt(self, name):
        return UInt64(func('visitParamExtractUInt', self._value, name))

    def visitParamExtractInt(self, name):
        return Int64(func('visitParamExtractInt', self._value, name))

    def visitParamExtractFloat(self, name):
        return Float64(func('visitParamExtractFloat', self._value, name))

    def visitParamExtractBool(self, name):
        return UInt8(func('visitParamExtractBool', self._value, name))

    def visitParamExtractRaw(self, name):
        return String(func('visitParamExtractRaw', self._value, name))

    def visitParamExtractString(self, name):
        return String(func('visitParamExtractString', self._value, name))

    has = visitParamHas
    getUInt = visitParamExtractUInt
    getInt = visitParamExtractInt
    getFloat = visitParamExtractFloat
    getBool = visitParamExtractBool
    getRaw = visitParamExtractRaw
    getString = visitParamExtractString

    def JSONHas(self, *indices_or_keys):
        return UInt8(func('JSONHas', self._value, *indices_or_keys))

    def JSONKey(self, key):
        return String(func('JSONKey', self._value, key))

    def JSONLength(self, *indices_or_keys):
        return UInt64(func('JSONLength', self._value, *indices_or_keys))

    def JSONType(self, *indices_or_keys):
        return String(func('JSONType', self._value, *indices_or_keys))

    def JSONExtractUInt(self, *indices_or_keys):
        return UInt64(func('JSONExtractUInt', self._value, *indices_or_keys))

    def JSONExtractInt(self, *indices_or_keys):
        return Int64(func('JSONExtractInt', self._value, *indices_or_keys))    
    
    def JSONExtractFloat(self, *indices_or_keys):
        return Float64(func('JSONExtractFloat', self._value, *indices_or_keys))

    def JSONExtractBool(self, *indices_or_keys):
        return UInt8(func('JSONExtractBool', self._value, *indices_or_keys))

    def JSONExtractString(self, *indices_or_keys):
        return String(func('JSONExtractString', self._value, *indices_or_keys))

    def JSONExtract(self, indices_or_keys, return_type):
        return TypeBase(Literal((func('JSONExtract', self._value, *indices_or_keys, return_type))))

    def JSONExtractKeysAndValues(self, *indices_or_keys, value_type):
        return TypeBase(Literal((func('JSONExtractKeysAndValues', self._value, *indices_or_keys, value_type))))

    def JSONExtractRaw(self, *indices_or_keys):
        return JSON(func('JSONExtractRaw', self._value, *indices_or_keys))



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import vulkn.engines
from vulkn import Vulkn
from vulkn.types.base import TypeBase, func, Literal
from vulkn.utils import timer


# TODO: Urgent - Fix before release. Add tests.


class ColumnVector(TypeBase):
    def __init__(self, value: any=None, name: str=None, n: str=None) -> None:
        if isinstance(value, Literal):
            self._value = value
        else:
            v = []
            for col in value:
                if isinstance(col, str):
                    v.append("'{}'".format(col))
                else:
                    v.append(str(col))
            self._value = Literal('SELECT arrayJoin([{}]) AS v'.format(','.join(v)))

    # Creation

    @classmethod
    def range(self, start, end, max_block_size=30000):
        q = f"""
            SELECT
                number AS i,
                number + {start} AS v 
            FROM numbers_mt(toUInt64(1 + abs({start} - {end})))
            SETTINGS max_block_size={max_block_size}"""
        return ColumnVector(Literal(q))

    @classmethod
    def rand(self, start, end, length, max_block_size=30000):
        q = f"""
            SELECT
                number AS i,
                {start} + rand64(number)%toUInt64(1 + abs({start} - {end})) AS v
            FROM numbers_mt({length})
            SETTINGS max_block_size={max_block_size}"""
        return ColumnVector(Literal(q))

    @classmethod
    def norm(self, mean, stddev, count):
        count = int(count)
        UInt32_MAX = vulkn.types.UInt32.MAX
        q = f"""
            SELECT
                rowNumberInAllBlocks() AS i,
                v
            FROM (
            SELECT
                arrayJoin(
                    arraySlice(
                        arrayReduce(
                            'groupArrayArray',
                            arrayMap(
                                i -> 
                                    [ ((sqrt(-2.0*log(rand(i)/{UInt32_MAX}))*cos(2*pi()*rand(i+100000000)/{UInt32_MAX}))*toFloat32({stddev}))+toFloat32({mean})
                                    , ((sqrt(-2.0*log(rand(i)/{UInt32_MAX}))*sin(2*pi()*rand(i+100000000)/{UInt32_MAX}))*toFloat32({stddev}))+toFloat32({mean})]
                                , range(toUInt64(ceil({count}/2))))), 1, {count})) AS v
            SETTINGS max_block_size=1)"""
        return ColumnVector(Literal(q))

    @classmethod
    def fromquery(self, query, column='v'):
        return ColumnVector(
            Literal(f"SELECT rowNumberInAllBlocks() AS i, {column} AS v FROM ({query})"))

    # Conversion

    def toArrayVector(self):
        from vulkn.types.array_vector import ArrayVector
        cache_table = self._cache()
        q = f"""
            SELECT
                groupArray(v) AS v
            FROM (
                SELECT v FROM {cache_table} WHERE i < 100000000 ORDER BY i LIMIT 100000000
            )"""
        return ArrayVector(Literal(q))

    # Manage

    def _cache(self, engine=None):
        ctx = Vulkn()
        engine = engine or vulkn.engines.Memory()
        return ctx.session.cache(self.sql, engine=engine)

    def alias(self, alias):
        self.vector_name = alias
        return self

    as_ = alias

    @timer
    def cache(self, engine=None):
        cache_table = self._cache(engine=engine)
        return ColumnVector(Literal(f'SELECT * FROM {cache_table}'))

    @timer
    def exec(self):
        ctx = Vulkn()
        return ctx._conn.select(self.sql)

    @timer
    def show(self, count=10, max_block_size=30000):
        ctx = Vulkn()
        sql = self.sql
        q = f'SELECT i, v FROM ({sql}) WHERE i <= {count} ORDER BY i LIMIT {count} SETTINGS max_block_size={max_block_size}'
        return ctx._conn.select(q).show()

    @timer
    def peek(self, count=10, max_block_size=30000):
        ctx = Vulkn()
        sql = self.sql
        q = f'SELECT i, v FROM ({sql}) WHERE i <= {count} ORDER BY i LIMIT {count} SETTINGS max_block_size={max_block_size}'
        return ctx._conn.select(q).to_records()

    def cast(self, to_type):
        vector = None
        toType = to_type if isinstance(to_type, str) else to_type.CAST
        if isinstance(self._value, list):
            vector = 'SELECT rowNumberInAllBlocks() AS i, arrayJoin({}) AS v'.format(str(self._value))
        else:
            vector = self.sql
        q = "SELECT i, cast(v, '{type}') AS v FROM ({vector})"
        return ColumnVector(Literal(q.format(type=toType, vector=vector)))

    # Operations

    def _cache_vector(self, cache):
        return self._cache() if cache else '({})'.format(str(self._value))

    def agg(self, agg):
        ctx = Vulkn()
        q = 'SELECT {agg}(v) AS {agg} FROM ({vector})'
        return ctx._conn.select(q.format(agg=agg, vector=self.sql)).to_records()

    def cut(self, cut_length, cache=False):
        vector = self._cache_vector(cache)
        q = f"""
            SELECT (i-i%{cut_length})/{cut_length} AS i, groupArray(v) AS v
            FROM ({vector})
            GROUP BY i ORDER BY i"""
        return ColumnVector(Literal(q))

    def delta(self, cache=False):
        vector = self._cache_vector(cache)
        q = f"SELECT i, runningDifference(v) AS v FROM (SELECT i, v FROM {vector} ORDER BY i)"
        return ColumnVector(Literal(q))

    def flatten(self, cache=False):
        vector = self._cache_vector(cache)
        q = f"""
            SELECT rowNumberInAllBlocks() AS i, _v AS v
            FROM (
                SELECT _v FROM ({vector}) 
                ARRAY JOIN arrayEnumerate(v) AS _i, v AS _v
                ORDER BY i, _i
            )"""
        return ColumnVector(Literal(q))

    def join(self, other, cache=False):
        vector = self._cache_vector(cache)
        other_vector = other._cache_vector(cache)
        q = f"""
            SELECT rowNumberInAllBlocks() AS i, v
            FROM (
                SELECT v FROM ({vector}) ORDER BY i
                UNION ALL
                SELECT v FROM ({other_vector}) ORDER BY i
            )"""
        return ColumnVector(Literal(q))

    def index_agg(self, agg):
        ctx = Vulkn()
        q = 'SELECT {agg}(i) AS {agg} FROM ({vector})'
        return ctx._conn.select(q.format(agg=agg, vector=self.sql)).to_records()    

    def limit(self, length, cache=False):
        vector = self._cache_vector(cache)
        return ColumnVector(Literal(f'SELECT * FROM ({vector}) LIMIT {length}'))

    def map(self, func, *args, cache=False):
        vector = self._cache_vector(cache)
        arg_list = '{},'.format(','.join(list(map(str, args)))) if len(args) > 0 else ''
        q = f'SELECT i, {func}({arg_list}v) AS v FROM ({vector})'
        return ColumnVector(Literal(q))

    def maplag(self, func, cache=False):
        vector = self._cache_vector(cache)
        q = f"""
            SELECT i, {func}(v1.v, v2.v) AS v FROM ({vector}) v1
            ALL INNER JOIN (
                SELECT toUInt64(i-1) AS i, v FROM ({vector})
            ) v2 USING (i)"""
        return ColumnVector(Literal(q))

    def maplead(self, func, cache=False):
        vector = self._cache_vector(cache)
        q = f"""
            SELECT i, {func}(v2.v, v1.v) AS v FROM ({vector}) v1
            ALL INNER JOIN (
                SELECT toUInt64(i-1) AS i, v FROM ({vector})
            ) v2 USING (i)"""
        return ColumnVector(Literal(q))

    def next(self, cache=False):
        vector = self._cache_vector(cache)
        q = f"""
            SELECT i, v FROM ({vector}) WHERE i > 0
            UNION ALL
            SELECT (SELECT COUNT() FROM {vector}), NULL"""
        return ColumnVector(Literal(q))

    def prev(self, cache=False):
        vector = self._cache_vector(cache)
        q = f"""
            SELECT 0 AS i, NULL AS v
            UNION ALL
            SELECT i+1, v FROM (
                SELECT * FROM ({vector}) WHERE i < (SELECT max(i-1) FROM ({vector}))
            )"""
        return ColumnVector(Literal(q))

    def reverse(self, cache=False):
        vector = self._cache_vector(cache)
        q = f"SELECT rowNumberInAllBlocks() AS i, v FROM (SELECT v FROM {vector} ORDER BY v DESC)"
        return ColumnVector(Literal(q))

    def shuffle(self, cache=False):
        vector = self._cache_vector(cache)
        q = f"SELECT rowNumberInAllBlocks() AS i, v FROM {vector} ORDER BY rand()"
        return ColumnVector(Literal(q))

    def sort(self, cache=False):
        vector = self._cache_vector(cache)
        q = f"SELECT rowNumberInAllBlocks() AS i, v FROM (SELECT v FROM {vector} ORDER BY v)"
        return ColumnVector(Literal(q))

    sort_reverse = reverse    

    def take(self, length):
        vector = None
        if isinstance(self._value, str) or isinstance(self._value, Literal):
            vector = 'SELECT groupArray(v) FROM (SELECT v FROM ({}) ORDER BY i)'.format(self.sql)
        else:
            vector = str(self)
        q = f"""
            SELECT rowNumberInAllBlocks() AS i, v
            FROM (
                WITH ({vector}) AS `#v`
                SELECT `#v`[(number%length(`#v`))+1] AS v
                FROM numbers_mt({length})
            )"""
        return ColumnVector(Literal(q))



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import random
import sys
import pprint
import argparse
import atexit
import logging


import vulkn
import vulkn.funcs as funcs
import vulkn.engines as engines
from vulkn.workspaces import UserWorkSpaceManager, LocalWorkSpace
from vulkn.funcs import and_, or_, not_, xor_, if_, multiIf, switch
from vulkn.types import *
from vulkn.utils import get_next_free_socket
from vulkn.datatable import WriteMode, JoinStrictness
from vulkn.config import VERSION


log = logging.getLogger()
log.setLevel('INFO')
ch = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
log.addHandler(ch)

parser = argparse.ArgumentParser(description='vulkn')

parser.add_argument('--list-folios', action='store_true')
parser.add_argument('--persist', action='store_true')
parser.add_argument('--local', action='store_true', default=False)
parser.add_argument('--save-auth', action='store_true')
parser.add_argument('--name', metavar='NAME', default=None)
parser.add_argument('--folio', metavar='FOLIO', default=None)
parser.add_argument('--workspace', metavar='WORKSPACE', default=None)
parser.add_argument('--host', metavar='HOST', default='localhost')
parser.add_argument('--user', metavar='USERNAME', default='default')
parser.add_argument('--password', metavar='PASSWORD', default='')
parser.add_argument('--port', metavar='PORT', default='9000')
parser.add_argument('--log-level', metavar='LOGLEVEL', default='INFO')
parser.add_argument('--timing', action='store_true')

args = parser.parse_args()

if args.list_folios:
    pprint.pprint(UserWorkSpaceManager().get_config())
    sys.exit(0)

if args.save_auth:
    if args.save_auth and args.server and args.user:
        UserWorkSpaceManager().save_auth(args.server, args.user)
    else:
        print('Authentication not set')
        sys.exit(1)
    sys.exit(0)

ce = None
if args.local:
    args.port = get_next_free_socket('127.0.0.1', list(range(9001,10000)))
    args.http_port = get_next_free_socket('127.0.0.1', list(range(8124,8999)))
    ce = LocalWorkSpace(persist=args.persist,
                        name=args.name,
                        workspace=args.workspace,
                        folio=args.folio,
                        port=args.port,
                        http_port=args.http_port)

v = vulkn.Vulkn(host=args.host, port=args.port, user=args.user, password=args.password)

vulkn.session.log.setLevel(args.log_level)
vulkn.session.timing = args.timing

tags = [
    "The environmentally friendly real-time analytics engine powered by ClickHouse.",
    "The developer friendly real-time analytics engine powered by ClickHouse.",
    "Stop waiting for your queries to complete and start having fun.",
    "ClickHouse - an analytics database for the 21st century."
]

print(f"""Copyright (C) 2019,2020 Jason Godden / VulknData Pty Ltd.

Добро пожаловать to VULKИ version {VERSION}!

██╗   ██╗██╗   ██╗██╗     ██╗  ██╗███╗   ██╗
██║   ██║██║   ██║██║     ██║ ██╔╝████╗  ██║
██║   ██║██║   ██║██║     █████╔╝ ██╔██╗ ██║
╚██╗ ██╔╝██║   ██║██║     ██╔═██╗ ██║╚██╗██║
 ╚████╔╝ ╚██████╔╝███████╗██║  ██╗██║ ╚████║
  ╚═══╝   ╚═════╝ ╚══════╝╚═╝  ╚═╝╚═╝  ╚═══╝                    
""")

print(tags[random.randint(0, len(tags)-1)])
print("\nVULKИ entrypoint initialized as 'v'.\n")
